# -- import packages: ---------------------------------------------------------
import neural_diffeqs
import torch

# -- import local dependencies: -----------------------------------------------
from . import base, mix_ins

# -- set type hints: ----------------------------------------------------------
from typing import List, Literal, Optional, Union

# -- DiffEq: ------------------------------------------------------------------
class LightningODE_FixedPotential(
    mix_ins.BaseForwardMixIn,
    mix_ins.PotentialMixIn,
    base.BaseLightningDiffEq,
):
    """LightningODE-FixedPotential"""
    def __init__(
        self,
        latent_dim: int = 50,
        name: Optional[str] = None,
        dt: float = 0.1,
        mu_hidden: Union[List[int], int] = [400, 400],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        train_lr=1e-4,
        train_optimizer=torch.optim.RMSprop,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        train_step_size: int = 10,
        backend: str = "auto",
        adjoint=False,
        # -- other: -----------------------------------------------------------
        loading_existing: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        LightningODE_FixedPotential

        Parameters:
        -----------
        latent_dim : int, optional
            Dimensionality of the latent space, by default 50
        name : str, optional
            Name of the model, by default None
        dt : float, optional
            Time step for the ODE solver, by default 0.1
        mu_hidden : Union[List[int], int], optional
            Hidden layer sizes for the neural network, by default [400, 400]
        mu_activation : Union[str, List[str]], optional
            Activation function(s) for the neural network, by default "LeakyReLU"
        mu_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the neural network, by default 0.2
        mu_bias : bool, optional
            Whether to use bias in the neural network, by default True
        mu_output_bias : bool, optional
            Whether to use bias in the output layer of the neural network, by default True
        mu_n_augment : int, optional
            Number of augmentations for the neural network, by default 0
        train_lr : float, optional
            Learning rate for training, by default 1e-4
        train_optimizer : torch.optim.Optimizer, optional
            Optimizer for training, by default torch.optim.RMSprop
        train_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for training, by default torch.optim.lr_scheduler.StepLR
        train_step_size : int, optional
            Step size for the learning rate scheduler, by default 10
        backend : str, optional
            Backend for the ODE solver, by default "auto"
        adjoint : bool, optional
            Whether to use the adjoint method for the ODE solver, by default False
        loading_existing : bool, optional
            Whether to load an existing model, by default False

        Returns:
        --------
        None

        Notes:
        ------
        This class implements a fixed potential ODE using PyTorch Lightning.

        Examples:
        ---------
        >>> model = LightningODE_FixedPotential(latent_dim=20, dt=0.05)
        >>> model.fit(data)
        """
        super().__init__()

        name = self._configure_name(name, loading_existing=loading_existing)

        self.save_hyperparameters()

        self._configure_torch_modules(func=neural_diffeqs.PotentialODE, kwargs=locals())
        self._configure_lightning_model(kwargs=locals())

    def __repr__(self) -> Literal['LightningODE-FixedPotential']:
        return "LightningODE-FixedPotential"



# -- import packages: ----------------------------------------------------------
import ABCParse
import pathlib
import anndata
import scanpy as sc
import os
import sklearn
import umap
import numpy as np
import web_files


# -- import local dependencies: ------------------------------------------------
from .. import io


# -- set typing: ---------------------------------------------------------------
from typing import Union


# -- Controller class: ---------------------------------------------------------
class PancreaticEndocrinogenesisDataset(ABCParse.ABCParse):
    HTTPS = "https://github.com/theislab/scvelo_notebooks/raw/master/data/Pancreas/endocrinogenesis_day15.h5ad"
    RAW_FNAME = "_downloaded.pancreas.h5ad"
    PROCESSED_FNAME = "pancreas.pp.h5ad"

    def __init__(self, data_dir=os.getcwd(), *args, **kwargs):
        self.__parse__(locals())

        if not self.data_dir.exists():
            self.data_dir.mkdir()

    @property
    def _scdiffeq_parent_data_dir(self):
        path = pathlib.Path(self._data_dir).joinpath("scdiffeq_data")
        if not path.exists():
            path.mkdir()
        return path

    @property
    def data_dir(self):
        path = self._scdiffeq_parent_data_dir.joinpath("pancreas")
        if not path.exists():
            path.mkdir()
        return path

    @property
    def raw_h5ad_path(self) -> pathlib.Path:
        return self.data_dir.joinpath(self.RAW_FNAME)

    @property
    def processed_h5ad_path(self) -> pathlib.Path:
        return self.data_dir.joinpath(self.PROCESSED_FNAME)

    def download_raw(self):
        web_file = web_files.WebFile(
            http_address=self.HTTPS, local_path=str(self.raw_h5ad_path)
        )
        web_file.download()

    @property
    def raw_adata(self):
        if not hasattr(self, "_raw_adata"):
            if not self.raw_h5ad_path.exists():
                self.download_raw()

            raw_adata = anndata.read_h5ad(self.raw_h5ad_path)
            self._raw_idx = raw_adata.obs.index

            del raw_adata.uns["pca"]
            del raw_adata.uns["neighbors"]
            del raw_adata.obsm["X_pca"]
            del raw_adata.obsm["X_umap"]
            del raw_adata.obsp["distances"]
            del raw_adata.obsp["connectivities"]

            raw_adata.layers["X_counts"] = raw_adata.X

            raw_adata.obs = raw_adata.obs.reset_index()
            raw_adata.obs.index = raw_adata.obs.index.astype(str)

            self._raw_adata = raw_adata

        return self._raw_adata

    @property
    def pp_adata(self):
        if not hasattr(self, "_pp_adata"):
            if not self.processed_h5ad_path.exists():
                self._pp_adata = self.get()
            else:
                self._pp_adata = anndata.read_h5ad(self.processed_h5ad_path)
        return self._pp_adata

    def pp(self, adata, min_genes=200, min_cells=3):

        adata = adata.copy()

        sc.pp.filter_cells(adata, min_genes=min_genes)
        sc.pp.filter_genes(adata, min_cells=min_cells)
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata)
        adata = adata[:, adata.var["highly_variable"]].copy()

        return adata

    def dimension_reduction(self, adata: anndata.AnnData):

        adata = adata.copy()

        self.SCALER_MODEL = sklearn.preprocessing.StandardScaler()
        self.PCA_MODEL = sklearn.decomposition.PCA(n_components=50)
        self.UMAP_MODEL = umap.UMAP(
            n_components=2, n_neighbors=30, random_state=0, min_dist=0.5
        )

        adata.obsm["X_scaled"] = self.SCALER_MODEL.fit_transform(adata.X.A)
        adata.obsm["X_pca"] = self.PCA_MODEL.fit_transform(adata.obsm["X_scaled"])
        adata.obsm["X_umap"] = self.UMAP_MODEL.fit_transform(adata.obsm["X_pca"])

        io.write_pickle(
            self.SCALER_MODEL, self.data_dir.joinpath("pancreas.scaler_model.pkl")
        )
        io.write_pickle(
            self.PCA_MODEL, self.data_dir.joinpath("pancreas.pca_model.pkl")
        )
        io.write_pickle(
            self.UMAP_MODEL, self.data_dir.joinpath("pancreas.umap_model.pkl")
        )

        return adata

    def annotate_time(self, adata):

        idx = adata.obs.index
        t0_str_idx = adata.obs[adata.obs["clusters"] == "Ductal"].index
        t0_idx = idx.isin(t0_str_idx)
        t1_idx = ~idx.isin(t0_str_idx)
        t = np.zeros(len(idx))
        t[t1_idx] = 1
        adata.obs["t"] = t

        return adata

    def get(self):

        adata_pp = self.pp(self.raw_adata)
        adata_pp = self.dimension_reduction(adata_pp)
        adata = self.annotate_time(adata_pp)
        adata.write_h5ad(self.processed_h5ad_path)
        return adata


def pancreas(data_dir: str = os.getcwd()) -> anndata.AnnData:
    """Pancreas dataset

    Parameters
    ----------
    data_dir : str, default=os.getcwd()
        Path to the directory where the data will be saved.

    Returns
    -------
    anndata.AnnData
        Processed AnnData object.
    """
    data_handler = PancreaticEndocrinogenesisDataset(data_dir=data_dir)
    return data_handler.pp_adata




from ._scverse_neighbors import SCVerseNeighbors, scverse_neighbors
from ._norm import L1Norm, L2Norm
from ._get_neighbor_indices import DistancesHandler, get_neighbor_indices
from ._compute_cosines import CosineCorrelation, ComputeCosines
from ._get_iterative_indices import get_iterative_indices
from ._transition_matrix import TransitionMatrix, transition_matrix



from ._mesh_grid import MeshGrid
from ._surface_3d import Surface3D
from ._neighbor_smoothing import NeighborSmoothing
from ._scatter_3d import Scatter3D
from ._potential_landscape import potential_landscape


# -- import packages: ----------------------------------------------------------
import ABCParse
import anndata
import logging
import os
import pathlib
import sklearn

# -- import local dependencies: ------------------------------------------------
from .. import io
from ._figshare_downloader import figshare_downloader

# -- configure logger: ----------------------------------------------------------
logger = logging.getLogger(__name__)


# -- Controller class: ---------------------------------------------------------
class LARRYInVitroDataset(ABCParse.ABCParse):
    FNAME = "larry.h5ad"
    figshare_id = 52612805

    def __init__(
            self,
            data_dir=os.getcwd(),
            filter_genes: bool = True,
            reduce_dimensions: bool = True,
            force_download: bool = False,
            *args,
            **kwargs,
        ):

        self.__parse__(locals())

    @property
    def _scdiffeq_parent_data_dir(self):
        path = pathlib.Path(self._data_dir).joinpath("scdiffeq_data")
        if not path.exists():
            path.mkdir()
        return path

    @property
    def data_dir(self):
        path = self._scdiffeq_parent_data_dir.joinpath("larry")
        if not path.exists():
            path.mkdir()
        return path

    @property
    def h5ad_path(self) -> pathlib.Path:
        return self.data_dir.joinpath(self.FNAME)

    @property
    def _DO_PREPROCESSING(self):
        return any([self._filter_genes, self._reduce_dimensions])

    def download(self):
        figshare_downloader(figshare_id=self.figshare_id, write_path=self.h5ad_path)

    def _gene_filtering(self, adata: anndata.AnnData) -> anndata.AnnData:
        return adata[:, adata.var["use_genes"]].copy()

    def _dimension_reduction(self, adata: anndata.AnnData):
        """Do sample dimension reduction"""
        # -- instantiate models: ----------------------------------------------
        scaler = sklearn.preprocessing.StandardScaler()
        pca = sklearn.decomposition.PCA(n_components=50)

        # -- fit transform data: ----------------------------------------------
        adata.obsm['X_scaled'] = scaler.fit_transform(adata.X.toarray())        
        adata.obsm['X_pca'] = pca.fit_transform(adata.obsm['X_scaled'])

        # -- save models: -----------------------------------------------------
        io.write_pickle(
            obj=scaler,
            path=self.data_dir.joinpath("scaler.pkl"),
        )
        io.write_pickle(
            obj=pca,
            path=self.data_dir.joinpath("pca.pkl"),
        )

    def _preprocess(self, adata: anndata.AnnData) -> anndata.AnnData:
        if self._DO_PREPROCESSING:
            logger.info("Preprocessing...")
            if self._filter_genes:
                adata = self._gene_filtering(adata)
            if self._reduce_dimensions:
                self._dimension_reduction(adata)
            adata.write_h5ad(self.h5ad_path)
        return adata

    @property
    def adata(self) -> anndata.AnnData:
        if not hasattr(self, "_adata"):
            if not self.h5ad_path.exists() or self._force_download:
                self.download()
                adata = anndata.read_h5ad(self.h5ad_path)
                self._adata = self._preprocess(adata=adata)
                return self._adata
            else:
                logger.info(f"Loading data from {self.h5ad_path}")
                return anndata.read_h5ad(self.h5ad_path)

def larry(
        data_dir: str = os.getcwd(),
        filter_genes: bool = True,
        reduce_dimensions: bool = True,
        force_download: bool = False,
    ) -> anndata.AnnData:
    """LARRY in vitro dataset

    Args:
        data_dir: str, default=os.getcwd()
            Path to the directory where the data will be saved.
        filter_genes: bool, default=True
            Whether to filter genes.
        reduce_dimensions: bool, default=True
            Whether to reduce dimensions.
        force_download: bool, default=False
            Whether to force download the data.

    Returns:
        anndata.AnnData: Preprocessed AnnData object.
    """
    data_handler = LARRYInVitroDataset(
        data_dir=data_dir,
        filter_genes=filter_genes,
        reduce_dimensions=reduce_dimensions,
        force_download=force_download,
    )
    return data_handler.adata



# -- import packages: ---------------------------------------------------------
import neural_diffeqs
import torch

# -- import local dependencies: -----------------------------------------------
from . import base, mix_ins

# -- set type hints: ----------------------------------------------------------
from typing import Literal, Optional, Union, List


# -- operational class: -------------------------------------------------------
class LightningSDE_VAE_FixedPotential(
    mix_ins.PreTrainMixIn,
    mix_ins.PotentialMixIn,
    mix_ins.VAEMixIn,
    base.BaseLightningDiffEq,
):
    """LightningSDE-VAE-FixedPotential"""
    def __init__(
        self,
        data_dim,
        latent_dim: int = 50,
        name: Optional[str] = None,
        mu_hidden: Union[List[int], int] = [400, 400],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_hidden: Union[List[int], int] = [400, 400],
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_dropout: Union[float, List[float]] = 0.2,
        sigma_bias: List[bool] = True,
        sigma_output_bias: bool = True,
        sigma_n_augment: int = 0,
        sde_type="ito",
        noise_type="general",
        brownian_dim=1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        backend="auto",
        train_lr=1e-5,
        train_optimizer=torch.optim.RMSprop,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        train_step_size=10,
        pretrain_lr=1e-3,
        pretrain_epochs=100,
        pretrain_optimizer=torch.optim.Adam,
        pretrain_scheduler=torch.optim.lr_scheduler.StepLR,
        pretrain_step_size=200,
        # -- encoder parameters: ----------------------------------------------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = "LeakyReLU",
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        # -- decoder parameters: ----------------------------------------------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = "LeakyReLU",
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        dt: float = 0.1,
        adjoint: bool = False,
        loading_existing: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        LightningSDE_VAE_FixedPotential

        Parameters
        ----------
        latent_dim : int, optional
            Dimensionality of the latent space, by default 50.
        name : str, optional
            Name of the model, by default None.
        mu_hidden : Union[List[int], int], optional
            Hidden layer sizes for the drift neural network, by default [400, 400].
        mu_activation : Union[str, List[str]], optional
            Activation function(s) for the drift neural network, by default 'LeakyReLU'.
        mu_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the drift neural network, by default 0.2.
        mu_bias : bool, optional
            Whether to use bias in the drift neural network, by default True.
        mu_output_bias : bool, optional
            Whether to use bias in the output layer of the drift neural network, by default True.
        mu_n_augment : int, optional
            Number of augmentations for the drift neural network, by default 0.
        sigma_hidden : Union[List[int], int], optional
            Hidden layer sizes for the diffusion neural network, by default [400, 400].
        sigma_activation : Union[str, List[str]], optional
            Activation function(s) for the diffusion neural network, by default 'LeakyReLU'.
        sigma_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the diffusion neural network, by default 0.2.
        sigma_bias : List[bool], optional
            Whether to use bias in the diffusion neural network, by default True.
        sigma_output_bias : bool, optional
            Whether to use bias in the output layer of the diffusion neural network, by default True.
        sigma_n_augment : int, optional
            Number of augmentations for the diffusion neural network, by default 0.
        sde_type : str, optional
            Type of stochastic differential equation, by default 'ito'.
        noise_type : str, optional
            Type of noise, by default 'general'.
        brownian_dim : int, optional
            Dimensionality of the Brownian motion, by default 1.
        coef_drift : float, optional
            Coefficient of drift, by default 1.0.
        coef_diffusion : float, optional
            Coefficient of diffusion, by default 1.0.
        backend : str, optional
            Backend for the SDE solver, by default "auto".
        train_lr : float, optional
            Learning rate for training, by default 1e-5.
        train_optimizer : torch.optim.Optimizer, optional
            Optimizer for training, by default torch.optim.RMSprop.
        train_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for training, by default torch.optim.lr_scheduler.StepLR.
        train_step_size : int, optional
            Step size for the training learning rate scheduler, by default 10.
        pretrain_lr : float, optional
            Learning rate for pretraining, by default 1e-3.
        pretrain_epochs : int, optional
            Number of epochs for pretraining, by default 100.
        pretrain_optimizer : torch.optim.Optimizer, optional
            Optimizer for pretraining, by default torch.optim.Adam.
        pretrain_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for pretraining, by default torch.optim.lr_scheduler.StepLR.
        pretrain_step_size : int, optional
            Step size for the pretraining learning rate scheduler, by default 200.
        encoder_n_hidden : int, optional
            Number of hidden layers in the encoder, by default 4.
        encoder_power : float, optional
            Power of the encoder, by default 2.
        encoder_activation : Union[str, List[str]], optional
            Activation function(s) for the encoder, by default 'LeakyReLU'.
        encoder_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the encoder, by default 0.2.
        encoder_bias : bool, optional
            Whether to use bias in the encoder, by default True.
        encoder_output_bias : bool, optional
            Whether to use bias in the output layer of the encoder, by default True.
        decoder_n_hidden : int, optional
            Number of hidden layers in the decoder, by default 4.
        decoder_power : float, optional
            Power of the decoder, by default 2.
        decoder_activation : Union[str, List[str]], optional
            Activation function(s) for the decoder, by default 'LeakyReLU'.
        decoder_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the decoder, by default 0.2.
        decoder_bias : bool, optional
            Whether to use bias in the decoder, by default True.
        decoder_output_bias : bool, optional
            Whether to use bias in the output layer of the decoder, by default True.
        dt : float, optional
            Time step for the SDE solver, by default 0.1.
        adjoint : bool, optional
            Whether to use the adjoint method for the SDE solver, by default False.
        loading_existing : bool, optional
            Whether to load an existing model, by default False.

        Returns
        -------
        None

        Notes
        -----
        This class implements a VAE with fixed potential SDE using PyTorch Lightning.

        Examples
        --------
        >>> model = LightningSDE_VAE_FixedPotential(data_dim=100, latent_dim=20, dt=0.05)
        >>> model.fit(data)
        """
        super().__init__()

        name = self._configure_name(name, loading_existing=loading_existing)

        self.save_hyperparameters()

        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func=neural_diffeqs.PotentialSDE, kwargs=locals())
        self._configure_lightning_model(kwargs=locals())

    def pretrain_step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X0_hat = self.Decoder(self.Encoder(batch.X0))
        recon_loss = self.reconstruction_loss(X0_hat, batch.X0).sum()
        self.log("pretrain_rl_mse", recon_loss.item())
        return recon_loss

    def training_step(self, batch, batch_idx, *args, **kwargs):
        if self.PRETRAIN:
            return self.pretrain_step(batch, batch_idx, stage="pretrain")
        return self.step(batch, batch_idx, stage="training")

    def __repr__(self) -> Literal['LightningSDE-VAE-FixedPotential']:
        return "LightningSDE-VAE-FixedPotential"




__module_name__ = "__init__.py"
__doc__ = """plotting __init__ module. Sub-package of the main scdiffeq API."""


# import functions accessed as sdq.pl.<func>: --------------------------------------------
# from .potential_landscape import potential_landscape
from ._velocity_stream import velocity_stream
# from ._smoothed_gene_expression import plot_smoothed_expression as smoothed_expression

# from ._temporal_expression import temporal_expression



from ._knn_mix_in import kNNMixIn
from ._state_characterization_mix_in import StateCharacterizationMixIn
from ._utility_io_mix_in import UtilityIOMixIn
from ._training_routine_mix_ins import PreTrainMixIn, TrainMixIn
from ._logging_mix_in import LoggingMixIn
from ._model_config_mix_in import ModelConfigMixIn



# -- import packages: --------------------------------------------------
import ABCParse
import adata_query
import anndata
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats
import sklearn.neighbors


# -- set typing: -------------------------------------------------------
from typing import Optional

import ABCParse
import numpy as np

# -----------------------------------------------------------------------------
# class Quiver(ABCParse.ABCParse):
#     """With autoscaling"""

#     def __init__(self, *args, **kwargs):
#         self.__parse__(locals())

#     @property
#     def _scale_factor(self):
#         """enables handling large values"""
#         return np.abs(self._X_emb).max()

#     def quiver(self, ax):
#         ax.quiver(
#             self._X_emb[:, 0] / self._scale_factor,
#             self._X_emb[:, 1] / self._scale_factor,
#             self._V_emb[:, 0],
#             self._V_emb[:, 1],
#             angles="xy",
#             scale_units="xy",
#             scale=None,
#         )

#     def __call__(self, ax, X_emb: np.ndarray, V_emb: np.ndarray):
#         """
#         X_emb (np.ndarray): state embedding.
#         V_emb (np.ndarray): velocity embedding.
#         """

#         self.__update__(locals())

#         self.quiver(ax)
   

# -- supporting function: --------------------------------------------------
def quiver_autoscale(X_emb: np.ndarray, V_emb: np.ndarray):
    """ """
    
    scale_factor = np.abs(X_emb).max()  # just so that it handles very large values
    fig, ax = plt.subplots()
    Q = ax.quiver(
        X_emb[:, 0] / scale_factor,
        X_emb[:, 1] / scale_factor,
        V_emb[:, 0],
        V_emb[:, 1],
        angles="xy",
        scale_units="xy",
        scale=None,
    )
    Q._init()
    fig.clf()
    plt.close(fig)
    return Q.scale / scale_factor


# -- operational class: -------------------------------------------------------
class GridVelocity(ABCParse.ABCParse):
    def __init__(
        self,
        density: float = 1,
        smooth: float = 0.5,
        n_neighbors: Optional[int] = None,
        min_mass: float = 1,
        autoscale=True,
        stream_adjust=True,
        cutoff_percentile: float = 0.05,
        *args,
        **kwargs
    ):
        
        self.__parse__(locals())

    @property
    def valid_idx(self):
        """use to rm invalid cells"""
        if not hasattr(self, "_valid_idx"):
            self._valid_idx = np.isfinite(self._X_emb.sum(1) + self._V_emb.sum(1))
        return self._valid_idx

    @property
    def X_emb(self):
        return self._X_emb[self.valid_idx]

    @property
    def V_emb(self):
        return self._V_emb[self.valid_idx]

    @property
    def n_obs(self):
        return self.X_emb.shape[0]

    @property
    def n_dim(self):
        return self.X_emb.shape[1]

    def _create_mesh_grid(self):
        """returns X_grid"""
        self._grs = []
        for dim_i in range(self.n_dim):
            m, M = np.min(self.X_emb[:, dim_i]), np.max(self.X_emb[:, dim_i])
            m = m - 0.01 * np.abs(M - m)
            M = M + 0.01 * np.abs(M - m)
            self._grs.append(np.linspace(m, M, int(50 * self._density)))

        return np.vstack([i.flat for i in np.meshgrid(*self._grs)]).T

    @property
    def n_neighbors(self):
        if not hasattr(self, "_n_neighbors") or self._n_neighbors is None:
            self._n_neighbors = int(self.n_obs / 50)
        return self._n_neighbors

    def _smooth_grid_neighbor_graph(self, X_emb, X_grid):
        """Fit ANN model to X_emb. Predict neighbors/dist for X_grid"""
        self._nn = sklearn.neighbors.NearestNeighbors(
            n_neighbors=self.n_neighbors, n_jobs=-1
        )
        self._nn.fit(X_emb)
        self._dists, self._neighbors = self._nn.kneighbors(X_grid)

    @property
    def scale(self):
        if not hasattr(self, "_scale"):
            self._scale = np.mean([(g[1] - g[0]) for g in self._grs]) * self._smooth
        return self._scale

    @property
    def weight(self):
        return scipy.stats.norm.pdf(x=self._dists, scale=self.scale)

    @property
    def p_mass(self):
        return self.weight.sum(1)

    def _estimate_grid_velocities(self, X_grid):
        self._smooth_grid_neighbor_graph(self.X_emb, X_grid)

        V_grid = (self.V_emb[self._neighbors] * self.weight[:, :, None]).sum(1)
        V_grid /= np.maximum(1, self.p_mass)[:, None]

        return V_grid

    def do_stream_adjust_X_grid(self, X_grid):
        return np.stack([np.unique(X_grid[:, 0]), np.unique(X_grid[:, 1])])

    def do_stream_adjust_V_grid(self, V_grid: np.ndarray) -> np.ndarray:
        """"""
        ns = int(np.sqrt(len(V_grid[:, 0])))
        V_grid = V_grid.T.reshape(2, ns, ns)
        mass = np.sqrt((V_grid**2).sum(0))
        min_mass = 10 ** (self._min_mass - 6)  # default min_mass = 1e-5
        min_mass = np.clip(min_mass, None, np.max(mass) * 0.9)

        cutoff = mass.reshape(V_grid[0].shape) < min_mass
        length = np.sum(np.mean(np.abs(self.V_emb[self._neighbors]), axis=1), axis=1).T
        length = length.reshape(ns, ns)
        cutoff |= length < np.percentile(length, self._cutoff_percentile)

        V_grid[0][cutoff] = np.nan

        return V_grid

    def __call__(self, X_emb: np.ndarray, V_emb: np.ndarray, *args, **kwargs):
        self.__update__(locals())

        X_grid = self._create_mesh_grid()
        V_grid = self._estimate_grid_velocities(X_grid)

        if self._stream_adjust:
            X_grid = self.do_stream_adjust_X_grid(X_grid)
            V_grid = self.do_stream_adjust_V_grid(V_grid)
        else:
            # filter on mass
            min_mass *= np.percentile(self.p_mass, 99) / 100
            X_grid, V_grid = X_grid[p_mass > min_mass], V_grid[p_mass > min_mass]

            if self._autoscale:
                V_grid /= 3 * quiver_autoscale(X_grid, V_grid)

        return X_grid, V_grid



import lightning
import pathlib
import pandas as pd

class LoggingMixIn(object):
    
    def _configure_CSVLogger(self):
        
        """ """
        
#         if hasattr(self, "_load_version"):
#             version = self._load_version
#         else:
#             version = None
                
        return lightning.pytorch.loggers.CSVLogger(
                save_dir=self._working_dir,
                name=self._name,
                version=None, # version,
                prefix="",
                flush_logs_every_n_steps=1,
            )
    
    @property
    def logger(self):
        if not hasattr(self, "_LOGGER"):
            self._csv_logger = self._configure_CSVLogger()
            if hasattr(self, "_logger") and (not self._logger is None):
                    self._LOGGER = [self._csv_logger, self._logger]
            else:
                self._LOGGER = [self._csv_logger]

        return self._LOGGER

    @property
    def version(self) -> int:
        return self.logger[0].version
    
    @property
    def _metrics_path(self):
        return pathlib.Path(self.logger[0].log_dir).joinpath("metrics.csv")
    
    @property
    def metrics(self):
        return pd.read_csv(self._metrics_path)




# -- import packages: ---------------------------------------------------------
import ABCParse
import torch
import numpy as np


# -- set typing: --------------------------------------------------------------
from typing import Union, List


class BatchRepresentation(object):
    """Representation of a batch of data.

    This class provides a textual representation of a batch of data, including its header,
    data tensor shapes, and fate index shape if available.

    Attributes:
        header (str): Header for the batch representation.
        X (str): Representation of the data tensor shape.
        W (str): Representation of the weight tensor shape.
        W_hat (str): Representation of the predicted weight tensor shape.
        F_idx (str): Representation of the fate index tensor shape if available, empty string otherwise.

    Methods:
        __call__: Generate the textual representation of the batch.

    """
    
    def __init__(self, batch):
        
        """Initialize the BatchRepresentation object.

        Args:
            batch: Batch of data.

        Returns:
            None

        """

        self.header = "🍪 LitDataBatch"
        self.X = f"- X: {list(batch.X.shape)}"
        self.W = f"- W: {list(batch.W.shape)}"
        self.W_hat = f"- W_hat: {list(batch.W_hat.shape)}"
        if not batch.F_idx is None:
            self.F_idx = f"- F_idx: {list(batch.F_idx.shape)}"
        else:
            self.F_idx = ""

    def __call__(self):
        """Generate the textual representation of the batch.

        Returns:
            str: Textual representation of the batch.

        """
        return "\n".join([self.header, self.X, self.W, self.W_hat, self.F_idx])


class BatchProcessor(ABCParse.ABCParse):    
    """Batch processor class for data preprocessing.

    This class provides methods for preprocessing a batch of data including sum normalization,
    and properties for accessing batch-related information.
    
    Batch Items:
    ------------
    0. t - time. always required
    1. X - data. tensor of size: [batch_size, len(t), n_dim]. always required.
    2. W - weight. tensor of size: [batch_size, len(t), 1]
    3. F_idx - fate_idx. tensor of size: [batch_Size, len(t), 1]

    Attributes:
        None

    Methods:
        _sum_normalize: Normalize the input tensor by sum along a specified axis.
        __repr__: Generate the textual representation of the batch.

    Properties:
        device: Get the device of the batch.
        n_batch_items: Get the number of items in the batch.
        batch_size: Get the batch size.
        t: Get unique time steps in the batch.
        X: Get the data tensor.
        X0: Get the initial data tensor.
        W_hat: Get the predicted weight tensor.
        W: Get the normalized weight tensor.
        F_idx: Get the fate index tensor if available.

    """

    def __init__(self, batch: List[torch.Tensor], batch_idx: int) -> None:
        """
        Initialize the BatchProcessor object.

        Args:
            batch (List[torch.Tensor]): List of tensors in the batch.
            batch_idx (int): Index of the batch.

        Returns:
            None

        """

        self.__parse__(locals(), private=["batch"])

    def _sum_normalize(
        self, X: Union[torch.Tensor, np.ndarray], sample_axis=1
    ) -> Union[torch.Tensor, np.ndarray]:
        """
        Normalize the input tensor by sum along a specified axis.

        Args:
            X (Union[torch.Tensor, np.ndarray]): Input tensor.
            sample_axis (int): Axis along which to perform normalization. Default is 1.

        Returns:
            Union[torch.Tensor, np.ndarray]: Normalized tensor.

        """
        
        return X / X.sum(sample_axis)[:, None]

    @property
    def device(self):
        """Get the device of the batch."""
        return self._batch[0].device

    @property
    def n_batch_items(self) -> int:
        """Get the number of items in the batch."""
        return len(self._batch)

    @property
    def batch_size(self) -> int:
        """Get the batch size."""
        return self._batch[0].shape[0]

    @property
    def t(self) -> torch.Tensor:
        """Get unique time steps in the batch."""
        return self._batch[0].unique()

    @property
    def X(self) -> torch.Tensor:
        """Get the cell state tensor."""
        return self._batch[1].transpose(1, 0).contiguous()

    @property
    def X0(self) -> torch.Tensor:
        """Get the initial cell state tensor."""
        return self.X[0]

    @property
    def W_hat(self) -> torch.Tensor:
        """Predicted weight tensor.

        If the predicted weight tensor is not cached, it will be computed and cached.
        """
        if not hasattr(self, "_W_hat"):
            if self.n_batch_items >= 3:
                W = self._batch[2].transpose(1, 0)
                for i in range(1, len(self.t)):
                    W[i] = torch.exp((self.t[i] - self.t[0]) * W[0])
                W[0] = torch.ones_like(W[0])
                self._W_hat = self._sum_normalize(W, sample_axis=1).contiguous()
            else:
                self._W_hat = self._sum_normalize(
                    torch.ones([len(self.t), self.batch_size, 1], device=self.device)
                ).contiguous()
        return self._W_hat

    @property
    def W(self) -> torch.Tensor:
        """Normalized weight tensor."""
        return self._sum_normalize(torch.ones_like(self.W_hat)).contiguous()

    @property
    def F_idx(self):
        """Fate index tensor if available."""
        if self.n_batch_items >= 4:
            return (
                self._batch[3]
                .transpose(1, 0)[0]
                .detach()
                .cpu()
                .numpy()
                .flatten()
                .astype(int)
                .astype(str)
            )

    def __repr__(self) -> str:
        """Generate the textual representation of the batch."""
        return BatchRepresentation(self)()



# -- import packages: ---------------------------------------------------------
import neural_diffeqs
import torch

# -- import local dependencies: -----------------------------------------------
from . import base, mix_ins

# -- set type hints: ----------------------------------------------------------
from typing import List, Literal, Optional, Union

# -- lightning model: ---------------------------------------------------------
class LightningSDE_PriorPotential(
    mix_ins.PotentialMixIn,
    mix_ins.DriftPriorMixIn,
    base.BaseLightningDiffEq,
):
    """LightningSDE-PriorPotential"""
    def __init__(
        self,
        latent_dim: int = 50,
        name: Optional[str] = None,
        train_lr=1e-4,
        train_optimizer=torch.optim.RMSprop,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        train_step_size=10,
        dt=0.1,
        adjoint=False,
        mu_hidden: Union[List[int], int] = [400, 400, 400],
        sigma_hidden: Union[List[int], int] = [400, 400, 400],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.2,
        sigma_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        sigma_bias: List[bool] = True,
        mu_output_bias: bool = True,
        sigma_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_n_augment: int = 0,
        sde_type="ito",
        noise_type="general",
        brownian_dim=1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        coef_prior_drift: float = 1.0,
        backend="auto",
        loading_existing: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        LightningSDE_PriorPotential

        Parameters
        ----------
        latent_dim : int, optional
            Dimensionality of the latent space, by default 50.
        name : str, optional
            Name of the model, by default None.
        train_lr : float, optional
            Learning rate for training, by default 1e-4.
        train_optimizer : torch.optim.Optimizer, optional
            Optimizer for training, by default torch.optim.RMSprop.
        train_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for training, by default torch.optim.lr_scheduler.StepLR.
        train_step_size : int, optional
            Step size for the training learning rate scheduler, by default 10.
        dt : float, optional
            Time step for the SDE solver, by default 0.1.
        adjoint : bool, optional
            Whether to use the adjoint method for the SDE solver, by default False.
        mu_hidden : Union[List[int], int], optional
            Hidden layer sizes for the drift neural network, by default [400, 400, 400].
        sigma_hidden : Union[List[int], int], optional
            Hidden layer sizes for the diffusion neural network, by default [400, 400, 400].
        mu_activation : Union[str, List[str]], optional
            Activation function(s) for the drift neural network, by default 'LeakyReLU'.
        sigma_activation : Union[str, List[str]], optional
            Activation function(s) for the diffusion neural network, by default 'LeakyReLU'.
        mu_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the drift neural network, by default 0.2.
        sigma_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the diffusion neural network, by default 0.2.
        mu_bias : bool, optional
            Whether to use bias in the drift neural network, by default True.
        sigma_bias : List[bool], optional
            Whether to use bias in the diffusion neural network, by default True.
        mu_output_bias : bool, optional
            Whether to use bias in the output layer of the drift neural network, by default True.
        sigma_output_bias : bool, optional
            Whether to use bias in the output layer of the diffusion neural network, by default True.
        mu_n_augment : int, optional
            Number of augmentations for the drift neural network, by default 0.
        sigma_n_augment : int, optional
            Number of augmentations for the diffusion neural network, by default 0.
        sde_type : str, optional
            Type of stochastic differential equation, by default 'ito'.
        noise_type : str, optional
            Type of noise, by default 'general'.
        brownian_dim : int, optional
            Dimensionality of the Brownian motion, by default 1.
        coef_drift : float, optional
            Coefficient of drift, by default 1.0.
        coef_diffusion : float, optional
            Coefficient of diffusion, by default 1.0.
        coef_prior_drift : float, optional
            Coefficient of prior drift, by default 1.0.
        backend : str, optional
            Backend for the SDE solver, by default "auto".
        loading_existing : bool, optional
            Whether to load an existing model, by default False.
        Returns
        -------
        None

        Notes
        -----
        This class implements a prior potential SDE using PyTorch Lightning.

        Examples
        --------
        >>> model = LightningSDE_PriorPotential(latent_dim=20, dt=0.05)
        >>> model.fit(data)
        """
        super().__init__()

        name = self._configure_name(name, loading_existing=loading_existing)

        self.save_hyperparameters()

        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func=neural_diffeqs.LatentPotentialSDE, kwargs=locals())
        self._configure_lightning_model(kwargs=locals())

    def __repr__(self) -> Literal['LightningSDE-PriorPotential']:
        return "LightningSDE-PriorPotential"



# import packages: ------------------------------------------------------------
import anndata


# API-facing function: --------------------------------------------------------
def read_h5ad(
    h5ad_path: str, silent: bool = False, annotate_path: bool = True
) -> anndata.AnnData:
    """Read an AnnData object from a path to a .h5ad file.

    Args:
        h5ad_path (str): Path to .h5ad file.
        silent (bool): If True, the AnnData object is not printed.
        annotate_path (bool): If True, adds the h5ad_path to the read AnnData object in adata.uns['h5ad_path'].

    Returns:
        anndata.AnnData
            The (annotated) single-cell data matrix of shape n_obs × n_vars. Rows correspond to cells and columns to genes.
        For more information, visit: https://anndata.readthedocs.io/en/latest/.

    Notes:
        Documentation for AnnData: https://anndata.readthedocs.io/en/stable/
    """

    adata = anndata.read_h5ad(h5ad_path)
    adata.uns["h5ad_path"] = h5ad_path

    if not silent:
        print(adata)

    return adata



# -- import packages: ---------------------------------------------------------
import neural_diffeqs
import torch

# -- import local dependencies: -----------------------------------------------
from . import mix_ins
from . import base

# -- set type hints: ----------------------------------------------------------
from typing import List, Literal, Optional, Union

# -- lightning model: ---------------------------------------------------------
class LightningSDE_VAE_PriorPotential(
    mix_ins.DriftPriorVAEMixIn,
    mix_ins.PotentialMixIn,
    mix_ins.VAEMixIn,
    mix_ins.PreTrainMixIn,
    base.BaseLightningDiffEq,
):
    def __init__(
        self,
        data_dim,
        latent_dim: int = 50,
        name: Optional[str] = None,
        train_lr=1e-5,
        pretrain_lr=1e-3,
        pretrain_epochs=100,
        pretrain_optimizer=torch.optim.Adam,
        train_optimizer=torch.optim.RMSprop,
        pretrain_scheduler=torch.optim.lr_scheduler.StepLR,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        pretrain_step_size=200,
        train_step_size=10,
        dt=0.1,
        adjoint=False,
        backend="auto",
        # -- sde params: ------------------------------------------------------
        mu_hidden: Union[List[int], int] = [400, 400, 400],
        sigma_hidden: Union[List[int], int] = [400, 400, 400],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.1,
        sigma_dropout: Union[float, List[float]] = 0.1,
        mu_bias: bool = True,
        sigma_bias: List[bool] = True,
        mu_output_bias: bool = True,
        sigma_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_n_augment: int = 0,
        sde_type="ito",
        noise_type="general",
        brownian_dim=1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        coef_prior_drift: float = 1.0,
        # -- encoder parameters: ----------------------------------------------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = "LeakyReLU",
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        # -- decoder parameters: ----------------------------------------------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = "LeakyReLU",
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        # -- other: ------------------------------------------------------------
        loading_existing: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        LightningSDE-VAE-PriorPotential model accessed as model.DiffEq

        Parameters
        ----------
        data_dim : int
            Dimensionality of the input data.
        latent_dim : int, optional
            Number of latent dimensions over which SDE should be parameterized. Default is 50.
        name : Optional[str], optional
            Model name used during project saving. Default is None.
        train_lr : float, optional
            Learning rate for training. Default is 1e-5.
        pretrain_lr : float, optional
            Learning rate for pretraining. Default is 1e-3.
        pretrain_epochs : int, optional
            Number of epochs for pretraining. Default is 100.
        pretrain_optimizer : torch.optim.Optimizer, optional
            Optimizer for pretraining. Default is torch.optim.Adam.
        train_optimizer : torch.optim.Optimizer, optional
            Optimizer for training. Default is torch.optim.RMSprop.
        pretrain_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for pretraining. Default is torch.optim.lr_scheduler.StepLR.
        train_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for training. Default is torch.optim.lr_scheduler.StepLR.
        pretrain_step_size : int, optional
            Step size for the pretraining learning rate scheduler. Default is 200.
        train_step_size : int, optional
            Step size for the training learning rate scheduler. Default is 10.
        dt : float, optional
            Time step for the SDE solver. Default is 0.1.
        adjoint : bool, optional
            Whether to use the adjoint method for backpropagation. Default is False.
        backend : str, optional
            Backend for the SDE solver. Default is "auto".
        mu_hidden : Union[List[int], int], optional
            Hidden layer sizes for the drift function. Default is [400, 400, 400].
        sigma_hidden : Union[List[int], int], optional
            Hidden layer sizes for the diffusion function. Default is [400, 400, 400].
        mu_activation : Union[str, List[str]], optional
            Activation function for the drift function. Default is "LeakyReLU".
        sigma_activation : Union[str, List[str]], optional
            Activation function for the diffusion function. Default is "LeakyReLU".
        mu_dropout : Union[float, List[float]], optional
            Dropout rate for the drift function. Default is 0.1.
        sigma_dropout : Union[float, List[float]], optional
            Dropout rate for the diffusion function. Default is 0.1.
        mu_bias : bool, optional
            Whether to use bias in the drift function. Default is True.
        sigma_bias : List[bool], optional
            Whether to use bias in the diffusion function. Default is True.
        mu_output_bias : bool, optional
            Whether to use bias in the output layer of the drift function. Default is True.
        sigma_output_bias : bool, optional
            Whether to use bias in the output layer of the diffusion function. Default is True.
        mu_n_augment : int, optional
            Number of augmented dimensions for the drift function. Default is 0.
        sigma_n_augment : int, optional
            Number of augmented dimensions for the diffusion function. Default is 0.
        sde_type : str, optional
            Type of SDE (e.g., "ito" or "stratonovich"). Default is "ito".
        noise_type : str, optional
            Type of noise (e.g., "general" or "diagonal"). Default is "general".
        brownian_dim : int, optional
            Dimension of the Brownian motion. Default is 1.
        coef_drift : float, optional
            Coefficient for the drift term. Default is 1.0.
        coef_diffusion : float, optional
            Coefficient for the diffusion term. Default is 1.0.
        coef_prior_drift : float, optional
            Coefficient for the prior drift term. Default is 1.0.
        encoder_n_hidden : int, optional
            Number of hidden layers for the encoder. Default is 4.
        encoder_power : float, optional
            Power for the encoder. Default is 2.
        encoder_activation : Union[str, List[str]], optional
            Activation function for the encoder. Default is "LeakyReLU".
        encoder_dropout : Union[float, List[float]], optional
            Dropout rate for the encoder. Default is 0.2.
        encoder_bias : bool, optional
            Whether to use bias in the encoder. Default is True.
        encoder_output_bias : bool, optional
            Whether to use bias in the output layer of the encoder. Default is True.
        decoder_n_hidden : int, optional
            Number of hidden layers for the decoder. Default is 4.
        decoder_power : float, optional
            Power for the decoder. Default is 2.
        decoder_activation : Union[str, List[str]], optional
            Activation function for the decoder. Default is "LeakyReLU".
        decoder_dropout : Union[float, List[float]], optional
            Dropout rate for the decoder. Default is 0.2.
        decoder_bias : bool, optional
            Whether to use bias in the decoder. Default is True.
        decoder_output_bias : bool, optional
            Whether to use bias in the output layer of the decoder. Default is True.
        loading_existing : bool, optional
            Whether to load an existing model. Default is False.

        Returns
        -------
        None
        """
        super().__init__()

        name = self._configure_name(name, loading_existing=loading_existing)

        self.save_hyperparameters()

        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(
            func=neural_diffeqs.LatentPotentialSDE, kwargs=locals()
        )
        self._configure_lightning_model(kwargs=locals())

    def __repr__(self) -> Literal['LightningSDE-VAE-PriorPotential']:
        return "LightningSDE-VAE-PriorPotential"



# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import adata_query
import logging
import numpy as np

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- controlling class: -------------------------------------------------------
class CellStateAnnotation(ABCParse.ABCParse):
    """Annotate cell states using a kNN graph."""

    def __init__(self, kNN: "kNN", silent: bool = False, *args, **kwargs) -> None:
        """Initialize `CellStateAnnotation` class.

        Args:
            kNN ('kNN')
                k-nearest neighbor graph.

        Returns:
            None
        """
        self.__parse__(locals())

    @property
    def _REF_DIM(self) -> int:
        """basis dim of kNN graph"""
        return self._kNN.X_use.shape[1]

    @property
    def _QUERY_DIM(self) -> int:
        """n_dim of cell query"""
        return self.X_query.shape[1]

    @property
    def X_query(self) -> np.ndarray:
        """Pull the simulated query cell data."""
        if not hasattr(self, "_X"):
            self._X = adata_query.fetch(
                adata=self._adata_sim, key=self._use_key, torch=False
            )
        return self._X

    def _assert_dimension_check(self) -> None:
        """Ensure the passed query dimension matches the basis dimension of the kNN graph."""

        msg = f"Query dim from {self._use_key}: {self._QUERY_DIM} does not match the kNN graph basis dimension: {self._REF_DIM}"

        assert self._QUERY_DIM == self._REF_DIM, msg

    @property
    def X_mapped(self) -> np.ndarray:
        """Map simulated query cells to observed cell state neighors."""
        if not hasattr(self, "_X_mapped"):
            self._X_mapped = self._kNN.aggregate(
                X_query=self.X_query, obs_key=self._obs_key, max_only=True
            )
        return self._X_mapped

    def forward(self) -> None:
        """Add mapped cell state values to adata_sim.obs"""
        self._adata_sim.obs[self._obs_key] = self.X_mapped.values.flatten()
        if not self._silent:
            logger.info(f"Added state annotation: adata_sim.obs['{self._obs_key}']")

    def __call__(
        self,
        adata_sim: anndata.AnnData,
        obs_key: str,
        use_key: str = "X",
        *args,
        **kwargs,
    ) -> None:
        """
        Args:
            adata_sim (anndata.AnnData)

            obs_key (str)

            use_key (str)

        Returns:
            None
        """

        self.__update__(locals())

        self._assert_dimension_check()
        self.forward()


# -- API-facing function: -----------------------------------------------------
def annotate_cell_state(
    adata_sim: anndata.AnnData,
    kNN: "kNN",
    obs_key: str = "state",
    use_key: str = "X",
    silent: bool = False,
) -> None:
    """
    Use a kNN Graph to annotate simulated cell states.

    Parameters
    ----------
    adata_sim : anndata.AnnData
        Simulated data object in the format of ``anndata.AnnData``, the (annotated)
        single-cell data matrix of shape ``n_obs × n_vars``. Rows correspond to cells
        and columns to genes. For more: [1](https://anndata.readthedocs.io/en/latest/).

    kNN : kNN
        k-nearest neighbor graph.

    obs_key : str, optional
        Observation key. Default is "state".

    use_key : str, optional
        Key to use for the basis. Default is "X".

    silent : bool, optional
        If True, suppresses informational messages. Default is False.

    Returns
    -------
    None

    References
    ----------
    .. [1] https://anndata.readthedocs.io/en/latest/
    """

    state_annot = CellStateAnnotation(kNN=kNN, silent=silent)
    state_annot(adata_sim=adata_sim, obs_key=obs_key, use_key=use_key)




# -- import packages: ----------------------------------------------------------
from lightning import Trainer
from lightning.pytorch import loggers
import torch
import os
import ABCParse


# -- import local dependencies: ------------------------------------------------
from ._lightning_callbacks_configuration import LightningCallbacksConfiguration
from .. import utils, callbacks


# -- define typing: ------------------------------------------------------------
from typing import Union, Dict, List, Optional


# -- Main class: ---------------------------------------------------------------
class LightningTrainerConfiguration(ABCParse.ABCParse):
    def __init__(self, save_dir: str = "scDiffEq_Model"):
        super().__init__()
        
        self.__parse__(locals())
        if not os.path.exists(self._save_dir):
            os.mkdir(self._save_dir)

    # -- kwargs: ---------------------------------------------------------------
#     @property
#     def _CSVLogger_kwargs(self):
#         return ABCParse.function_kwargs(
#             func=loggers.CSVLogger, kwargs=self._PARAMS, ignore=['version'],
#         )

    @property
    def _Trainer_kwargs(self):
        ignore=["accelerator", "callbacks", 'version', 'logger']
        tk = ABCParse.function_kwargs(
            func=Trainer,
            kwargs=self._PARAMS,
        )
        for val in ignore:
            if val in tk.keys():
                tk.pop(val)
        return tk

    @property
    def Callbacks(self):
        callback_config = LightningCallbacksConfiguration()
        
        return callback_config(
            version = self._version,
            monitor_hardware=self._monitor_hardware,
            stage = self._stage,
            viz_frequency = self._viz_frequency,
            model_name=self._model_name,
            working_dir=self._working_dir,
            train_version=self._train_version,
            pretrain_version=self._pretrain_version,
            callbacks=self._callbacks,
            ckpt_frequency=self._ckpt_frequency,
            keep_ckpts=self._keep_ckpts,
            monitor=self._monitor,
            retain_test_gradients=self._retain_test_gradients,
            save_last=self._save_last_ckpt,
            # swa_lrs=1e-5,
        )

#         return callback_config(
#             callbacks=self._callbacks,
#             ckpt_frequency=self.ckpt_frequency,
#             keep_ckpts=self.keep_ckpts,
#             retain_test_gradients=self.retain_test_gradients,
#             monitor = self.monitor,
# #             swa_lrs = self.swa_lrs,
#             save_last = self.save_last_ckpt,
#             version = self.version,
#         )
    
    @property
    def accelerator(self):
        if not self._accelerator is None:
            return self._accelerator
        if torch.cuda.is_available():
            return "gpu"
        if torch.backends.mps.is_available(): # experimental
            return "mps"
        return "cpu"

    # -- trainers: -------------------------------------------------------------
    @property
    def Trainer(self):
        """
        Main Lightning Trainer used for fitting / testing.
        If pre-train routine was used, Trainer loads from ckpt path.
        """

        return Trainer(
            accelerator=self.accelerator,
            logger=self._logger, # loggers.CSVLogger(**self._CSVLogger_kwargs),
            callbacks=self.Callbacks,
            **self._Trainer_kwargs,
        )

    @property
    def GradientsRetainedTestTrainer(self):
        """
        Quasi test trainer - serves as a workaround for evaluating test data
        while retaining gradients.
        """

        self._Trainer_kwargs["max_epochs"] = 0
        self._Trainer_kwargs["callbacks"] = self.Callbacks

        return Trainer(
            accelerator=self.accelerator,
            logger=self._logger,# loggers.CSVLogger(**self._CSVLogger_kwargs),
            num_sanity_val_steps=-1,
            enable_progress_bar=False,
            **self._Trainer_kwargs,
        )

    def __call__(
        self,
        logger,
        lr: float = None,
        model_name="scDiffEq_model",
        gradient_clip_val: float = 0.5,
        monitor_hardware: bool = False,
        working_dir=os.getcwd(),
        train_version=0,
        pretrain_version=0,
        viz_frequency = 1,
        stage=None,
        max_epochs=500,
        monitor=None,
        accelerator=None,
        devices=1,
        prefix: str = "",
        log_every_n_steps=1,
        flush_logs_every_n_steps: int = 1,
        ckpt_frequency: int = 25,
        save_last_ckpt: bool = True,
        keep_ckpts: int = -1,
        version: Optional[Union[int, str]] = None,
        callbacks: list = [],
        potential_model: bool = False,
        check_val_every_n_epoch = 1,
        limit_val_batches = None,
        num_sanity_val_steps = None,
        val_check_interval = None,
#         swa_lrs: float = None,
        reload_dataloaders_every_n_epochs = 1,
        **kwargs
    ):
        
        """
        Return trainer upon call.
        
        Parameters:
        -----------
        stage
            type: str
            default: None
        
        max_epochs
            type: int
            default: 500
        
        accelerator
            type: str
            default: None
        
        devices
            type: int
            default: None
        
        prefix
            type: str
            default: ""
        
        log_every_n_steps
            type: int
            default: 1
        
        flush_logs_every_n_steps
            type: int
            default: 1
        
        version
            type: [int, str]
            default: None
        
        callbacks
            type: list
            default: []
        
        potential_model
            type: bool
            default: False
            
        kwargs:
        -------
        Keyword arguments that may be passed to pytorch_lightning.Trainer()
        
        Notes:
        ------
        """
        
        self._gradient_clip_val = gradient_clip_val
        self._logger = logger

        self._retain_test_gradients = False
        if stage is None:
            stage = ""
        self._stage = stage
#         if isinstance(swa_lrs, NoneType):
#             swa_lrs = lr
            
        if torch.cuda.device_count() > 0:
            devices = torch.cuda.device_count()

        self.__parse__(locals()) #  private=['accelerator', 'callbacks'])
        
#         self._PARAMS["name"] = "{}_logs".format(stage)
#         log_save_dir = os.path.join(self.save_dir, self._PARAMS["name"])
#         if not os.path.exists(log_save_dir):
#             os.mkdir(log_save_dir)

        if (potential_model) and (stage in ["test", "predict"]):
            self._retain_test_gradients = True
            return self.GradientsRetainedTestTrainer

        return self.Trainer




__module_name__ = "__init__.py"
__doc__ = """tools __init__ module. Sub-package of the main scdiffeq API."""
__author__ = ", ".join(["Michael E. Vinyard",])
__email__ = ", ".join(["mvinyard.ai@gmail.org",])


# import functions accessed as sdq.tl.<func>: ---------------------------------

from . import utils


# -- kNN: ---------------------------------------------------------------------
from ._knn import kNN
from ._knn_smoothing import kNNSmoothing


# -- simulation: --------------------------------------------------------------
from ._simulation import Simulation, simulate


# -- post-simulation annotation functions: ------------------------------------
from ._annotate_gene_features import GeneCompatibility, annotate_gene_features
from ._annotate_cell_state import annotate_cell_state
from ._annotate_cell_fate import annotate_cell_fate
from ._invert_scaled_gex import invert_scaled_gex


# -- perturbation: ------------------------------------------------------------
from ._fate_perturbation_experiment import FatePerturbationExperiment
from ._fate_perturbation_screen import FatePerturbationScreen
from ._perturb import perturb
from ._perturb_scan_z_range import perturb_scan_z_range


# -- velocity plotting tools: -------------------------------------------------
from ._grid_velocity import GridVelocity
from ._velocity_embedding import VelocityEmbedding
from ._velocity_graph import velocity_graph



# -- general characterization: ------------------------------------------------
from ._drift_diffusion_state_characterization import drift, diffusion
from ._cell_potential import cell_potential, normalize_cell_potential

# -- --------------------------------------------------------------------------
# from ._feature_correlation import (
#     FeatureCorrelation,
#     drift_correlated_features,
#     diffusion_correlated_features,
#     potential_correlated_features,
# )

# from ._grouped_expression import GroupedExpression, grouped_expression
# from ._temporal_expression import temporal_expression






# -- import packages: ---------------------------------------------------------
import sklearn.preprocessing
import numpy as np
import lightning
import logging
import anndata
import torch
import adata_query
import logging

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- import local dependencies: -----------------------------------------------
from ..core import utils
from ._knn_smoothing import kNNSmoothing

# from ._fetch import fetch
from ._knn import kNN
import ABCParse

# -- set typing: --------------------------------------------------------------
from typing import Union

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- operator classes: --------------------------------------------------------
class CellPotentialNormalization(ABCParse.ABCParse):
    """
    Procedure occurs in 5 steps:
    1. Sign flip
    2. Clip outliers via quantile cutoff
    3. kNN Smoothing
    4. Log-transformation
    5. MinMaxScaling

    Notes:
    ------
    1.  Log-transform results in slightly better per-cell
        correlatation w/ CytoTRACE

    """

    def __init__(
        self,
        q: float = 0.05,
        raw_psi_key: str = "_psi",
        kNN_use_key: str = "X_pca",
        knn_smoothing_iters: int = 5,
        use_tqdm: bool = True,
    ) -> None:

        """
        Parameters:
        -----------
        q: quantile cutoff
        """

        self.__parse__(locals(), public=[None])

    @property
    def kNN(self) -> kNN:
        if not hasattr(self, "_graph"):
            self._graph = kNN(self.adata, use_key=self._kNN_use_key)
        return self._graph

    @property
    def _RAW_PSI(self):
        return self.adata.obs[self._raw_psi_key].values

    @property
    def _POSITIVE_RAW_PSI(self):
        """STEP 1: Sign flip"""
        return self._RAW_PSI * -1

    @property
    def _Q_MIN(self):
        return self._q

    @property
    def _Q_MAX(self):
        return 1 - self._q

    @property
    def _Q_MIN_CUTOFF(self):
        return np.quantile(self._POSITIVE_RAW_PSI, self._Q_MIN)

    @property
    def _Q_MAX_CUTOFF(self):
        return np.quantile(self._POSITIVE_RAW_PSI, self._Q_MAX)

    def _clip_psi(self):
        """STEP 2: Clip outlier values using a quantile"""
        logger.info("Quantile-clipping outliers")
        _CLIPPED_PSI = self._POSITIVE_RAW_PSI.copy()
        _CLIPPED_PSI[self._POSITIVE_RAW_PSI < self._Q_MIN_CUTOFF] = self._Q_MIN_CUTOFF
        _CLIPPED_PSI[self._POSITIVE_RAW_PSI > self._Q_MAX_CUTOFF] = self._Q_MAX_CUTOFF
        return _CLIPPED_PSI

    @property
    def _CLIPPED_PSI(self):
        """STEP 2: Clip outlier values using a quantile"""
        if not hasattr(self, "_clipped_psi"):
            self._clipped_psi = self._clip_psi()
        return self._clipped_psi

    def _knn_smoothing(self):
        """STEP 3: kNN smoothing"""
        self.adata.obs["_CLIPPED_PSI"] = self._CLIPPED_PSI
        smoothing = kNNSmoothing(
            self.adata,
            kNN=self.kNN,
            use_key=self._kNN_use_key,
            n_iters=self._knn_smoothing_iters,
            use_tqdm=self._use_tqdm,
        )
        logger.info("kNN smoothing")
        return smoothing(key="_CLIPPED_PSI", add_to_adata=False)

    @property
    def _SMOOTHED_PSI(self):
        """STEP 3: kNN smoothing"""
        if not hasattr(self, "_smoothed_psi"):
            self._smoothed_psi = self._knn_smoothing()
        return self._smoothed_psi

    @property
    def _LOG_PSI(self):
        """STEP 4: Log-transform"""
        logger.info("Log-transforming")
        if not hasattr(self, "_log_psi"):
            self._log_psi = np.log10(self._SMOOTHED_PSI)
        return self._log_psi

    def _min_max_scaling(self):
        """STEP 5: min-max scaling"""
        logger.info("Scaling")
        scaler = sklearn.preprocessing.MinMaxScaler()
        return scaler.fit_transform(self._LOG_PSI.reshape(-1, 1))

    @property
    def _SCALED_PSI(self):
        """STEP 5: min-max scaling"""
        if not hasattr(self, "_scaled_psi"):
            self._scaled_psi = self._min_max_scaling()
        return self._scaled_psi

    def _clean_up_adata(self) -> None:
        self.adata.obs.drop("_CLIPPED_PSI", axis=1, inplace=True)

    def __call__(self, adata, key_added="psi") -> None:
        
        self.__update__(locals(), public = ['adata'])
        adata.obs[key_added] = self._SCALED_PSI
        self._clean_up_adata()


class CellPotential(ABCParse.ABCParse):
    """Calculate [raw] potential values, given cells and a model."""

    def __init__(
        self,
        use_key: str = "X_pca",
        device: Union[str, torch.device] = torch.device("cuda:0"),
        seed: int = 0,
        gpu = True,
    ) -> None:
        
        """ """

        self.__parse__(locals(), public=[None])
        lightning.seed_everything(0)

    @property
    def Z_input(self):
        if not hasattr(self, "_Z_input"):
            self._Z_input = adata_query.fetch(
                adata=self.adata,
                key=self._use_key,
                torch=self._gpu,
                device=self._device,
            )
        return self._Z_input

    def forward(self, model):
        logger.info("Computing")
        return model.DiffEq.DiffEq.mu(self.Z_input).flatten().detach().cpu().numpy()

    def __call__(
        self, adata: anndata.AnnData, model, key_added: str = "_psi"
    ) -> None:

        self.__update__(locals(), public=["adata"])
        self.Z_psi = self.forward(model)
        adata.obs[key_added] = self.Z_psi


# -- API-facing functions: -----------------------------------------------------
def normalize_cell_potential(
    adata,
    use_key="_psi",
    key_added="psi",
    q=0.05,
    kNN_use_key="X_pca",
    knn_smoothing_iters=5,
    use_tqdm=True,
) -> None:
    """Can be AnnData from a simulation or the original AnnData object containing observed cells"""

    cell_potential_norm = CellPotentialNormalization(
        q=q,
        raw_psi_key=use_key,
        kNN_use_key=kNN_use_key,
        knn_smoothing_iters=knn_smoothing_iters,
        use_tqdm=use_tqdm,
    )

    cell_potential_norm(adata, key_added=key_added)


def cell_potential(
    adata: anndata.AnnData,
    model,
    use_key: str = "X_pca",
    raw_key_added: str = "_psi",
    norm_key_added: str = "psi",
    device: Union[str, torch.device] = torch.device("cuda:0"),
    seed: int = 0,
    normalize: bool = True,
    return_raw_array: bool = False,
    q: float = 0.05,
    knn_smoothing_iters: int = 5,
    use_tqdm: bool = True,
    gpu: bool = True,
):
    """ """
    cell_potential = CellPotential(use_key=use_key, device=device, seed=seed, gpu=gpu)
    cell_potential(adata=adata, model=model, key_added=raw_key_added)

    if normalize:
        normalize_cell_potential(
            adata=adata,
            use_key=raw_key_added,
            key_added=norm_key_added,
            q=q,
            kNN_use_key=use_key,
            knn_smoothing_iters=knn_smoothing_iters,
            use_tqdm=use_tqdm,
        )
    if return_raw_array:
        return cell_potential.Z_psi




from lightning import Callback
from datetime import datetime
import pickle
import os

class Testing(Callback):
    def __init__(self):
        self.now = str(datetime.now())

    def _mk_output_dir(self, trainer):
        self.dpath = os.path.join(trainer.logger.log_dir, "model_test_outputs")
        if not os.path.exists(self.dpath):
            os.mkdir(self.dpath)
            
    def on_test_batch_end(self, trainer, model, outputs, batch, batch_idx, dataloader_idx, *args, **kwargs):
        
        """ """
        
        self._mk_output_dir(trainer)
        
        fpath = os.path.join(self.dpath, "test_outputs.{}.pkl".format(self.now))
        pickle.dump(obj=outputs,
                    file=open(fpath, "wb"),
                    protocol=pickle.HIGHEST_PROTOCOL,
                   )




import pandas as pd
import numpy as np

from ._mesh_grid import MeshGrid
from ...core import utils
import ABCParse

class Surface3D(ABCParse.ABCParse):
    """
    Mesh grid surface.
    
    Uses a passed mesh grid and pre-fit neighbor model to map
    the points of a mesh grid to the corresponding y-val.
    """

    def __init__(self, gridpoints: int = 250):

        self.__parse__(locals(), public=[None])

    def _build_mesh_grid(self):
        self._mesh_grid = MeshGrid(
            adata=self.adata, gridpoints=self._gridpoints
        )
        self._x, self._y = self._mesh_grid()
        return self._mesh_grid

    @property
    def X(self):
        return self._x.reshape(-1, 1)

    @property
    def Y(self):
        return self._y.reshape(-1, 1)

    @property
    def mesh_grid_coordinates(self):
        return np.concatenate([self.X, self.Y], axis=1)
    
    def _to_frame(self, z_pred):
        self._surface_df = pd.DataFrame(
            data=z_pred, index=self._mesh_grid._X_GRID, columns=self._mesh_grid._Y_GRID
        )
        return self._surface_df
        

    def predict_from_model(self):
        return self.neighbor_model.predict(self.mesh_grid_coordinates).reshape(
            self._gridpoints, self._gridpoints
        )

    def __call__(self, adata, neighbor_model):
        """ """
        
        self.__update__(locals())
        self._build_mesh_grid()
        
        return self._to_frame(self.predict_from_model())




import ABCParse
import anndata

from .utils import ComputeCosines, scverse_neighbors


from typing import Dict, Optional


# -- API-facing function: ------------------------------------------------
def velocity_graph(
    adata: anndata.AnnData,
    state_key: str = "X_pca",
    velocity_key: str = "X_drift",
    n_pcs: Optional[int] = None,
    velocity_key_added: str = "velocity",
    split_negative: bool = True,
    silent: bool = False,
    neighbor_kwargs: Dict = {},
    *args,
    **kwargs,
):
    """
    Args:
        adata (anndata.AnnData)
        
        state_key (str): ... **Default** = "velocity"
        
        velocity_key (str): ... **Default** = "X_drift"
        
        velocity_key_added (str): ... **Default** = "velocity"
        
        split_negative (bool): ... **Default** = True
        
        silent (bool): **Default** = False.
        
        neighbor_kwargs (Optional[Dict[str, Any]]): **Default** = {},
        
    
    Returns:
        (None)
    """
    
    scverse_neighbors(adata, silent=silent, **neighbor_kwargs)
    
    init_kw = ABCParse.function_kwargs(ComputeCosines.__init__, locals())
    call_kw = ABCParse.function_kwargs(ComputeCosines.__call__, locals())

    compute_cosines = ComputeCosines(**init_kw)
    compute_cosines(**call_kw)




# -- import packages: ---------------------------------------------------------
import ABCParse
import logging
import os


# -- import local dependencies: -----------------------------------------------
from .. import lightning_models, utils


# -- set typing: --------------------------------------------------------------
from typing import Dict, Optional


# -- setup logging: -----------------------------------------------------------
logger = logging.getLogger(__name__)


# -- operational class: -------------------------------------------------------
class LightningModelConfiguration(ABCParse.ABCParse):
    """ """
    _potential_types = {
        "fixed": "FixedPotential",
        "prior": "PriorPotential",
    }

    def __init__(
        self,
        data_dim,
        latent_dim: int = 50,
        DiffEq_type: str = "SDE",
        potential_type: str = "fixed",
        fate_bias_csv_path = None,
        velocity_ratio_params: Dict = None,
    ):
        """ """

        self.__parse__(locals(), public=[None])

    @property
    def available_lightning_models(self):
        return [
            attr
            for attr in lightning_models.__dir__()
            if attr.startswith("Lightning")
        ]

    @property
    def DiffEq_type(self):
        return "".join(["Lightning", self._DiffEq_type])

    @property
    def use_vae(self):
        return self._data_dim > self._latent_dim

    @property
    def potential_type(self):
        if self._potential_type:
            return self._potential_types[self._potential_type]
        
    @property
    def fate_bias_aware(self) -> bool:
        """
        First checks if the past is passed. If so, checks if it exists.
        If passed but path does not exist, raises a ValueError.
        """
        if not self._fate_bias_csv_path is None:
            if not os.path.exists(self._fate_bias_csv_path):
                raise ValueError("Path to fate_bias.csv was passed though not found.")
            else:
                return True
        else:
            return False
        
    @property
    def _USE_CKPT(self) -> bool:
        return not self._ckpt_path is None
    
    def __call__(
        self, kwargs: Dict, ckpt_path: Optional[str] = None, loading_existing: bool = False,
                ):
        
        self._ckpt_path = ckpt_path

        _model = [self.DiffEq_type]

        if self.use_vae:
            _model.append("VAE")

        if self.potential_type:
            _model.append(self.potential_type)
            
        if self.fate_bias_aware:
            _model.append("FateBiasAware")
            
            
        if self._velocity_ratio_params:
            _model.append("RegularizedVelocityRatio")

        _model = "_".join(_model)
        
        if _model in self.available_lightning_models:
            lit_model = getattr(lightning_models, _model)
            
            if self._USE_CKPT:
                return lit_model.load_from_checkpoint(self._ckpt_path)
            
            model_kwargs = utils.function_kwargs(func=lit_model.__init__, kwargs=kwargs)
            model_kwargs['loading_existing'] = loading_existing
            logger.warning(model_kwargs)
            return lit_model(**model_kwargs) # data_dim = self._data_dim, 
        
        raise ValueError(f"Configuration tried: {_model} - this does not exist as an available model.")



from collections import OrderedDict
import torch_nets
import torch

import ABCParse

# -- lower-level class: -------
class ComponentStateLoader(ABCParse.ABCParse):
    """
    A lower level class, meant to interact with specific model components.
    Used by FlexibleComponentLoader.
    """

    def __init__(self, component, *args, **kwargs):

        self._COMPONENT = component

    @property
    def _EXPECTED_LOAD_MSG(self):
        return "<All keys matched successfully>"

    @property
    def _MODEL_COMPONENT(self) -> OrderedDict:
        return getattr(self._model.DiffEq, self._COMPONENT)

    @property
    def _STATE(self):
        return self._MODEL_COMPONENT.state_dict()

    def _transfer(self, ckpt):

        TransferStateDict = {}

        for key in ckpt["state_dict"].keys():
            if self._COMPONENT in key:
                new_key = key.split(f"{self._COMPONENT}.")[1]
                TransferStateDict[new_key] = ckpt["state_dict"][key]

        return OrderedDict(TransferStateDict)

    @property
    def ckpt(self):
        return torch.load(self._ckpt_path, map_location=self._map_location)

    def plot_wandb(self):
        torch_nets.pl.weights_and_biases(self._STATE)

    def __check__(self):
        if not str(self._msg) == self._EXPECTED_LOAD_MSG:
            raise UserWarning("Check keys - there might have been a problem.")

    def __call__(self, model, ckpt_path, map_location=None):

        self.__parse__(locals(), public=[None])

        ckpt_state = self._transfer(self.ckpt)
        self._msg = self._MODEL_COMPONENT.load_state_dict(ckpt_state)

        self.__check__()


# ---- API-facing class: ---------------------------------------------
class FlexibleComponentLoader(ABCParse.ABCParse):
    """
    Load individual component state dicts. for example just VAE or just DiffEq. Or all of it.
    """
    def __init__(self, model):

        self.__parse__(locals())

    def load_encoder_state(self, ckpt_path):

        self._encoder_ckpt_path = ckpt_path

        self.encoder_loader = ComponentStateLoader("Encoder")
        self.encoder_loader(self.model, self._encoder_ckpt_path)

    def load_decoder_state(self, ckpt_path):
        """"""
        self._decoder_ckpt_path = ckpt_path
        self.decoder_loader = ComponentStateLoader("Decoder")
        self.decoder_loader(self.model, self._decoder_ckpt_path)

    def load_DiffEq_state(self, ckpt_path):
        """"""
        self._diffeq_ckpt_path = ckpt_path
        self.diffeq_loader = ComponentStateLoader("DiffEq")
        self.diffeq_loader(self.model, self._diffeq_ckpt_path)

    def load_VAE(self, ckpt_path):
        """"""
        self.load_encoder_state(ckpt_path)
        self.load_decoder_state(ckpt_path)




from types import FunctionType
from typing import Optional


import ABCParse
import anndata
import autodevice
import torch


from ... import tools


class StateCharacterizationMixIn(object):
    """MixIn container for state characterization functions."""

    def __check_adata(self, adata: anndata.AnnData) -> anndata.AnnData:
        """Check if adata was passed."""
        if adata is None:
            return self.adata
        return adata

    def __check_use_key(self, use_key: str) -> str:
        """Check if adata was passed."""
        if use_key is None:
            return self._use_key
        return use_key

    def __check_vital_inputs(self, adata: anndata.AnnData, use_key: str):
        """Check both adata and use_key"""
        adata = self.__check_adata(adata)
        use_key = self.__check_use_key(use_key)

        return adata, use_key

    def _drift_diffusion_fwd(
        self,
        func: FunctionType,
        obsm_key_added: str,
        obs_key_added: str,
        adata: Optional[anndata.AnnData] = None,
        use_key: Optional[str] = None,
        device=autodevice.AutoDevice(),
        inplace: bool = True,
        silent: bool = False,
        *args,
        **kwargs,
    ):
        """"""
        adata, use_key = self.__check_vital_inputs(adata, use_key)

        return func(
            adata=adata,
            model=self,
            use_key=use_key,
            obsm_key_added=obsm_key_added,
            obs_key_added=obs_key_added,
            device=device,
            inplace=inplace,
            silent=silent,
            *args,
            **kwargs,
        )

    def drift(
        self,
        adata: Optional[anndata._core.anndata.AnnData] = None,
        use_key: Optional[str] = None,
        obsm_key_added: str = "X_drift",
        obs_key_added: str = "drift",
        device=autodevice.AutoDevice(),
        inplace: bool = True,
        silent: bool = False,
        *args,
        **kwargs,
    ):
        """Drift method."""
        self._drift_diffusion_fwd(
            func=tools.drift,
            obsm_key_added=obsm_key_added,
            obs_key_added=obs_key_added,
            adata=adata,
            use_key=use_key,
            device=device,
            inplace=inplace,
            silent=silent,
            *args,
            **kwargs,
        )
    

    def diffusion(
        self,
        adata: Optional[anndata._core.anndata.AnnData] = None,
        use_key: Optional[str] = None,
        obsm_key_added: str = "X_diffusion",
        obs_key_added: str = "diffusion",
        device=autodevice.AutoDevice(),
        inplace: bool = True,
        silent: bool = False,
        *args,
        **kwargs,
    ):
        """Diffusion method."""

        self._drift_diffusion_fwd(
            func=tools.diffusion,
            obsm_key_added=obsm_key_added,
            obs_key_added=obs_key_added,
            adata=adata,
            use_key=use_key,
            device=device,
            inplace=inplace,
            silent=silent,
            *args,
            **kwargs,
        )

    def potential(
        self,
        use_key: str = "X_pca",
        raw_key_added: str = "_psi",
        norm_key_added: str = "psi",
        seed: int = 0,
        normalize: bool = True,
        return_raw_array: bool = False,
        q: float = 0.05,
        knn_smoothing_iters: int = 5,
        device: Optional[torch.device] = autodevice.AutoDevice(),
        use_tqdm: bool = True,
    ):
        """Potential method."""

        tools.cell_potential(
            adata=self.adata,
            model=self,
            use_key=use_key,
            raw_key_added=raw_key_added,
            norm_key_added=norm_key_added,
            device=device,
            seed=seed,
            normalize=normalize,
            return_raw_array=return_raw_array,
            q=q,
            knn_smoothing_iters=knn_smoothing_iters,
            use_tqdm=use_tqdm,
        )





import autodevice

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import os
import lightning
import scdiffeq_plots as sdq_pl
from tqdm.notebook import tqdm


from .. import utils


from typing import Dict
NoneType = type(None)
import ABCParse



class VisualizePredictions(lightning.Callback, ABCParse.ABCParse):
    def __init__(
        self,
        reducer,
        kNN,
        adata,
        use_key,
        t,
        t0_idx,
        N=200,
        frequency=1,
        outdir=None,
        device=autodevice.AutoDevice(),
        label_cmap=None,
        label_key = None,
        simulation_cmap = matplotlib.cm.tab20.colors,
        *args,
        **kwargs,
    ):

        # will have to come from model.reducer.X_umap in the future (could also come from adata)
        X_umap = reducer.X_umap

        self.__parse__(locals(), private=[None])
        self._configure_time(t)
        

    def _configure_time(self, t):
        self.t = t.to(self.device)
        self.t_cmap = sdq_pl.temporal_cmap(t=t)
    
    @property
    def X0(self):
        return utils.fetch_format(
            self.adata, use_key=self.use_key, idx=self.t0_idx, N=self.N
        )

    def _FORWARD(self, pl_module, X0, t):

        X_hat = {}
        X_hat_kNN = {}
        X_hat_UMAP = {}
        for i in tqdm(range(len(self.t0_idx)), desc="Simulating", leave=False):
            key = self.t0_idx[i]
            X_hat[key] = xh = pl_module(X0[i], self.t)
            X_hat_kNN[key] = self.kNN(
                xh, annot_key=self.label_key, query_t=None
            )
            X_hat_UMAP[key] = np.stack(
                [
                    self.reducer.UMAP.transform(xh_j)
                    for xh_j in xh.detach().cpu().numpy()
                ]
            )
        return X_hat, X_hat_kNN, X_hat_UMAP

    def _background(self):

        xu = self.X_umap
        fig, axes = sdq_pl.plot(nplots=2, ncols=2, rm_ticks=True, delete_spines=["all"])

        for ax in axes:
            ax.scatter(xu[:, 0], xu[:, 1], c="lightgrey", s=2, alpha=0.5, zorder=0)

        return fig, axes

    def _label(self, X_hat_kNN, idx):
        # if you want a single label, uncomment below
        # .value_counts().idxmax()
        return X_hat_kNN[idx].idxmax(1)

    @property
    def SAVE_PATH(self):
        fname = f"./sample_model_output.epoch_{self.epoch}.png"
        if not isinstance(self.outdir, NoneType):
            return os.path.join(self.outdir, fname)
        return fname

    def _color_outline_by_simulation(self, ax, xu_x, xu_y, i, t, sizes):
        ax.scatter(
                    xu_x,
                    xu_y,
                    color=self.simulation_cmap[i],
                    zorder=t + 1,
                    s=sizes[0],
                    ec="None",
                )
        ax.scatter(
                    xu_x,
                    xu_y,
                    c="w",
                    zorder=t + 2,
                    s=sizes[1],
                    ec="None",
                )
    
    def plot(self, X_hat_umap: Dict, X_hat_kNN, sizes=[80, 40, 20]):

        fig, axes = self._background()

        for i, (idx, X_umap) in enumerate(X_hat_umap.items()):
            for t in range(len(X_umap)):
                if i == 0:
                    label = str(round(self.t[t].item(), 3))
                else:
                    label = None
                if t == 0:
                    xu_x, xu_y = X_umap[t][:, 0].mean(), X_umap[t][:, 1].mean()
                else:
                    xu_x, xu_y = X_umap[t][:, 0], X_umap[t][:, 1]

                self._color_outline_by_simulation(axes[0], xu_x, xu_y, i, t, sizes)
                self._color_outline_by_simulation(axes[1], xu_x, xu_y, i, t, sizes)
                
                axes[0].scatter(
                    xu_x,
                    xu_y,
                    color=self.t_cmap[t],
                    zorder=t + 3,
                    s=sizes[2],
                    ec="None",
                    label=label,
                )
                if not isinstance(self.label_cmap, NoneType):
                    color = X_hat_kNN[idx][t].map(self.label_cmap)
                    if t == 0:
                        color = color[0]
                else:
                    color = "dimgrey"

                axes[1].scatter(
                    xu_x, xu_y, c=color, ec="None", zorder=t + 3
                )
        if len(self.t) <= 5:
            axes[0].legend(edgecolor="None", facecolor="None")
            
            
        return fig, axes

    def __call__(self, DiffEq, t=None, sizes=[80, 40, 20], resimulate=True):

        if not isinstance(t, NoneType):
            self._configure_time(t)

        X_hat, X_hat_kNN, X_hat_umap = self._FORWARD(
            DiffEq.to(self.device), self.X0, self.t
        )
        fig, axes = self.plot(X_hat_umap, X_hat_kNN, sizes=sizes)
        plt.savefig(self.SAVE_PATH)
        plt.close()

    def on_fit_start(self, trainer, pl_module, *args, **kwargs):        
        self.epoch = pl_module.current_epoch
        
        self.__call__(pl_module)

    def on_train_epoch_end(self, trainer, pl_module, *args, **kwargs):
        self.epoch = pl_module.current_epoch
        
        if self.epoch % self.frequency == 0:
            self.__call__(pl_module)        



# -- import packages: ---------------------------------------------------------
import pathlib
import pickle


# -- set typing: --------------------------------------------------------------
from typing import Any, Dict, Optional, Union


# -- operational class: -------------------------------------------------------
class PickleIO:
    def __init__(self, *args, **kwargs):
        """ """
        pass

    def read(self, path, mode="rb"):
        return pickle.load(self.__path__(path, mode))

    def write(self, obj, path, mode="wb", protocol=pickle.HIGHEST_PROTOCOL):
        """If writing for use in colab, use protocol=4"""

        pickle.dump(obj=obj, file=self.__path__(path, mode), protocol=protocol)

    def __path__(self, path, mode):
        return open(path, mode)


# -- API-facing functions: ----------------------------------------------------
def read_pickle(path: Union[str, pathlib.Path], mode: Optional[str] = "rb"):
    """Read the contents of a pickle file into memory.

    Args:
        path (Union[str, pathlib.Path]): pickle file path.

        mode (Optional[str]): read mode. **Default** = "rb"

    Returns:
        obj (Dict[Any,Any]): Object, usually a dictionary contained in pickle file.
    """

    pickle_io = PickleIO()
    return pickle_io.read(path, mode=mode)


def write_pickle(
    obj: Dict[Any, Any],
    path: Union[str, pathlib.Path],
    mode: Optional[str] = "wb",
    protocol: Optional[int] = pickle.HIGHEST_PROTOCOL,
) -> None:
    """Save an object to a pickle file.

    Args:
        obj (Dict[Any, Any]): Object, usually a dictionary to write to pickle file.

        path (Union[str, pathlib.Path]): Path to which pickle file should be written.

        mode (Optional[str]): Write mode. **Default** = "wb"

        protocol (Optional[int]): Pickling protocol. **Default** = pickle.HIGHEST_PROTOCOL

    Returns:
        None
    """

    pickle_io = PickleIO()
    pickle_io.write(obj=obj, path=path, mode=mode, protocol=protocol)



# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import cellplots as cp
import logging
import matplotlib.cm
import matplotlib.pyplot as plt
import numpy as np
import os
import pathlib

# -- import local dependencies: -----------------------------------------------
from ..tools import VelocityEmbedding, GridVelocity

# -- set type hints: ----------------------------------------------------------
from typing import Any, Dict, List, Optional, Union, Tuple

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- Operational class: -------------------------------------------------------
class VelocityStreamPlot(ABCParse.ABCParse):
    """A class to generate and plot velocity stream plots on a given plt.Axes object.

    This class is designed to visualize the flow of cells (or particles) in a
    velocity field, commonly used in single-cell data analysis to represent cell
    trajectories over a reduced dimensionality space. It extends the ABCParse class
    to leverage its parsing capabilities for initializing and configuring the plot
    parameters.

    Attributes:
        density (float): The density of the grid for velocity vectors. Defaults to 1.
        smooth (float): Smoothing factor applied to the velocity field. Defaults to 0.5.
        n_neighbors (Optional[int]): Number of neighbors to consider for local averaging. Defaults to None.
        min_mass (float): Minimum mass (weight) threshold for considering a point in the velocity field. Defaults to 1.
        autoscale (bool): Flag to automatically scale the vectors. Defaults to True.
        stream_adjust (bool): Adjust the streamplot parameters for optimal visualization. Defaults to True.
        cutoff_percentile (float): Percentile for cutoff to ignore outlier velocities. Defaults to 0.05.
        velocity_key (str): Key in `adata` to access velocity vectors. Defaults to "velocity".
        self_transitions (bool): Whether to consider self-transitions in the velocity calculations. Defaults to True.
        use_negative_cosines (bool): Flag to use negative cosines to adjust directionality. Defaults to True.
        T_scale (float): Scaling factor for the transition matrix. Defaults to 10.
        args: Variable length argument list.
        kwargs: Arbitrary keyword arguments.

    Note:
        This class requires an AnnData object `adata` to be passed at call time, not at initialization.
    """

    def __init__(
        self,
        density: float = 1,
        smooth: float = 0.5,
        n_neighbors: Optional[int] = None,
        min_mass: float = 1,
        autoscale: bool = True,
        stream_adjust: bool = True,
        cutoff_percentile: float = 0.05,
        velocity_key: str = "velocity",
        self_transitions: bool = True,
        use_negative_cosines: bool = True,
        T_scale: float = 10,
        *args,
        **kwargs,
    ) -> None:
        """Initializes the VelocityStreamPlot object with parameters to configure the velocity stream plot.

        This method sets up the necessary parameters for generating a velocity stream plot, including the setup for velocity embedding and grid velocity calculation. It parses the arguments and initializes internal states needed for plotting.

        Args:
            density (float): Density of the grid for velocity vectors. Higher values create a denser grid. Defaults to 1.
            smooth (float): Smoothing factor applied to the velocity vectors, influencing the smoothness of the stream plot. Defaults to 0.5.
            n_neighbors (Optional[int]): Number of nearest neighbors to use for local averaging of velocities. If None, a default heuristic is used. Defaults to None.
            min_mass (float): Minimum mass (weight) threshold for considering a point in the velocity field. Helps in filtering out noise. Defaults to 1.
            autoscale (bool): If True, scales the magnitude of velocity vectors automatically based on the density and size of the plot. Defaults to True.
            stream_adjust (bool): If True, adjusts stream plot parameters for optimal visualization. Defaults to True.
            cutoff_percentile (float): Percentile for cutoff to filter out outlier velocities, specified as a fraction between 0 and 1. Defaults to 0.05.
            velocity_key (str): Key in the AnnData object `adata` to access velocity vectors. Defaults to "velocity".
            self_transitions (bool): If True, considers self-transitions in velocity calculations, affecting the direction and magnitude of vectors. Defaults to True.
            use_negative_cosines (bool): If True, uses negative cosines to adjust the directionality of vectors, potentially improving visualization clarity. Defaults to True.
            T_scale (float): Scaling factor for the transition matrix T, affecting the overall magnitude of velocity vectors. Defaults to 10.
            args: Additional positional arguments not specifically defined.
            kwargs: Additional keyword arguments not specifically defined.

        Note:
            The `__init__` method does not require the AnnData object `adata`. Instead, `adata` should be passed to the `__call__` method when generating the plots.
        """

        self.__parse__(locals())

        self._velocity_emb = VelocityEmbedding(
            velocity_key=velocity_key,
            self_transitions=self_transitions,
            use_negative_cosines=use_negative_cosines,
            T_scale=T_scale,
        )
        self._grid_velocity = GridVelocity(
            density=density,
            smooth=smooth,
            n_neighbors=n_neighbors,
            min_mass=min_mass,
            autoscale=autoscale,
            stream_adjust=stream_adjust,
            cutoff_percentile=cutoff_percentile,
        )

    @property
    def X_emb(self):
        if not hasattr(self, "_X_emb"):
            self._X_emb, self._V_emb = self._velocity_emb(self._adata)
        return self._X_emb

    @property
    def V_emb(self):
        if not hasattr(self, "_V_emb"):
            self._X_emb, self._V_emb = self._velocity_emb(self._adata)
        return self._V_emb

    @property
    def X_grid(self):
        if not hasattr(self, "_X_grid"):
            self._X_grid, self._V_grid = self._grid_velocity(self.X_emb, self.V_emb)
        return self._X_grid

    @property
    def V_grid(self):
        if not hasattr(self, "_V_grid"):
            self._X_grid, self._V_grid = self._grid_velocity(self.X_emb, self.V_emb)
        return self._V_grid

    @property
    def x(self):
        return self.X_grid[0]

    @property
    def y(self):
        return self.X_grid[1]

    @property
    def u(self):
        return self.V_grid[0]

    @property
    def v(self):
        return self.V_grid[1]

    @property
    def xmin(self):
        return np.min(self.X_emb[:, 0])

    @property
    def xmax(self):
        return np.max(self.X_emb[:, 0])

    @property
    def ymin(self):
        return np.min(self.X_emb[:, 1])

    @property
    def ymax(self):
        return np.max(self.X_emb[:, 1])

    @property
    def xmargin(self):
        return (self.xmax - self.xmin) * self._add_margin

    @property
    def ymargin(self):
        return (self.ymax - self.ymin) * self._add_margin

    def _set_margin(self, ax):
        """"""
        ax.set_xlim(self.xmin - self.xmargin, self.xmax + self.xmargin)
        ax.set_ylim(self.ymin - self.ymargin, self.ymax + self.ymargin)

    @property
    def _STREAMPLOT_KWARGS(self) -> Dict[str, Any]:
        kwargs = {
            "color": self._stream_color,
            "density": self._stream_density,
            "linewidth": self._linewidth,
            "zorder": self._stream_zorder,
            "arrowsize": self._arrowsize,
            "arrowstyle": self._arrowstyle,
            "maxlength": self._maxlength,
            "integration_direction": self._integration_direction,
        }
        kwargs.update(self._stream_kwargs)
        return kwargs

    def streamplot(self, ax) -> None:
        ax.streamplot(self.x, self.y, self.u, self.v, **self._STREAMPLOT_KWARGS)
        self._set_margin(ax)

    @property
    def _SCATTER_KWARGS(self) -> Dict[str, Any]:
        """ """
        kwargs = {
            "c": self._c,
            "zorder": self._scatter_zorder,
            "ec": "None",
            "alpha": 0.2,
            "s": 50,
            "cmap": self._cmap,
            "rasterized": self._rasterized,
        }
        kwargs.update(self._scatter_kwargs)
        return kwargs

    def _SCATTER_CMAP(self, groups) -> Dict:
        """ """
        if not hasattr(self, "_cmap"):
            self._cmap = matplotlib.cm.tab20.colors
        if not isinstance(self._cmap, Dict):
            self._cmap = {group: self._cmap[en] for en, group in enumerate(groups)}
        return self._cmap

    def scatter(self, ax) -> None:
        """Generates a scatter plot on the given matplotlib axis, overlaying the stream plot.

        This method visualizes individual points (cells) on the velocity stream plot,
        with optional coloring and grouping.

        Args:
            ax (matplotlib.axes.Axes): The matplotlib axis on which to plot the scatter plot.
        """
        obs_df = self._adata.obs.copy().reset_index()
        cols = obs_df.columns.tolist()

        kwargs = self._SCATTER_KWARGS

        COLOR_FROM_OBS = self._c in cols

        if COLOR_FROM_OBS:
            COLOR_BY_GROUP = str(obs_df[self._c].dtype) == "categorical"
            if not COLOR_BY_GROUP:  # implies float not grouped object.
                c_idx = np.argsort(obs_df[self._c])
                kwargs.update({"c": obs_df[self._c][c_idx]})
                self._img = ax.scatter(
                    self.X_emb[c_idx, 0], self.X_emb[c_idx, 1], **kwargs
                )
                if not self._disable_cbar:
                    cbar = plt.colorbar(mappable=self._img, **self._cbar_kwargs)
                    cbar.solids.set(alpha=1)
            #                     cbar.set_alpha(1)

            else:
                kwargs.pop("c")
                groups = obs_df.groupby(self._c).groups  # dict
                cmap = self._SCATTER_CMAP(groups)
                for group, group_ix in groups.items():
                    if hasattr(self, "_group_zorder") and group in self._group_zorder:
                        kwargs.update({"zorder": self._group_zorder[group]})
                    ax.scatter(
                        self.X_emb[group_ix, 0],
                        self.X_emb[group_ix, 1],
                        color=cmap[group],
                        **kwargs,
                    )
        else:
            ax.scatter(self.X_emb[:, 0], self.X_emb[:, 1], **kwargs)

    @property
    def scdiffeq_figure_dir(self):
        return pathlib.Path("scdiffeq_figures")

    def _mk_fig_dir(self):
        if not self.scdiffeq_figure_dir.exists():
            os.mkdir(self.scdiffeq_figure_dir)
            logger.info(f"mkdir: {self.scdiffeq_figure_dir}")

    @property
    def sdq_info(self):
        return self._adata.uns["sdq_info"]

    @property
    def data_model_info_tag(self) -> str:
        return f"{self.sdq_info['project']}.version_{self.sdq_info['version']}.ckpt_{self.sdq_info['ckpt']}"

    @property
    def fname_basis(self) -> pathlib.Path:
        if "sdq_info" in self._adata.uns:
            try:
                return self.scdiffeq_figure_dir.joinpath(
                    f"velocity_stream.{self.data_model_info_tag}"
                )
            except:
                return self.scdiffeq_figure_dir.joinpath(
                    f"velocity_stream.{self.sdq_info}"
                )
            finally:
                pass
        return self.scdiffeq_figure_dir.joinpath("velocity_stream")

    @property
    def SVG_path(self) -> pathlib.Path:
        """ """
        return pathlib.Path(".".join([str(self.fname_basis), "svg"]))

    @property
    def PNG_path(self) -> pathlib.Path:
        return pathlib.Path(".".join([str(self.fname_basis), "png"]))

    def save_img(self) -> None:
        """Saves the generated plot to both SVG and PNG formats in a specified directory."""
        self._mk_fig_dir()
        plt.savefig(self.SVG_path, dpi=self._svg_dpi)
        plt.savefig(self.PNG_path, dpi=self._png_dpi)
        logger.info(f"Saved to: \n  {self.SVG_path}\n  {self.PNG_path}")

    def __call__(
        self,
        adata: anndata.AnnData,
        ax: Optional[Union[plt.Axes, List[plt.Axes]]] = None,
        stream_color: str = "k",
        c: str = "dodgerblue",
        group_zorder: Optional[Dict] = None,
        cmap: Optional[Union[Dict, List, Tuple, str]] = "plasma_r",
        linewidth: float = 0.5,
        stream_density: float = 2.5,
        add_margin: float = 0.1,
        arrowsize: float = 1,
        density: float = 1,
        arrowstyle: str = "-|>",
        maxlength: float = 4,
        integration_direction: str = "both",
        scatter_zorder: int = 0,
        stream_zorder: int = 10,
        rasterized: bool = True,
        mpl_kwargs: Optional[Dict] = {},
        scatter_kwargs: Optional[Dict] = {},
        stream_kwargs: Optional[Dict] = {},
        cbar_kwargs: Optional[Dict] = {},
        disable_scatter: bool = False,
        disable_cbar: bool = False,
        save: bool = False,
        png_dpi: Optional[float] = 500,
        svg_dpi: Optional[float] = 250,
        *args,
        **kwargs,
    ):
        """Generates velocity stream plots for the provided AnnData object.

        Args:
            adata (anndata.AnnData): The AnnData object containing the data for plotting.
            ax (Optional[Union[plt.Axes, List[plt.Axes]]]): A matplotlib axis or list of axes where plots will be drawn.
            **kwargs: Additional keyword arguments to customize the plot appearance.

        Returns:
            List[plt.Axes]: A list of matplotlib axes with the generated plots.
        """
        self.__update__(locals())

        if ax is None:
            _mpl_kwargs = {
                "nplots": 1,
                "ncols": 1,
                "height": 1.0,
                "width": 1.0,
                "delete": "all",
                "del_xy_ticks": [True],
            }
            _mpl_kwargs.update(mpl_kwargs)
            fig, axes = cp.plot(**_mpl_kwargs)

        else:
            axes = ABCParse.as_list(ax)

        for ax in axes:
            self.streamplot(ax)
            if not self._disable_scatter:
                self.scatter(ax)

        if self._save:
            self.save_img()

        return axes


# -- API-facing function: -----------------------------------------------------
def velocity_stream(
    adata: anndata.AnnData,
    ax: Optional[Union[plt.Axes, List[plt.Axes]]] = None,
    c: str = "dodgerblue",
    cmap: Optional[Union[Dict, List, Tuple]] = "plasma_r",
    group_zorder: Optional[Dict] = None,
    linewidth: float = 0.5,
    stream_density: float = 2.5,
    add_margin: float = 0.1,
    arrowsize: float = 1,
    arrowstyle: str = "-|>",
    maxlength: float = 4,
    integration_direction: str = "both",
    scatter_zorder: int = 101,
    stream_zorder: int = 201,
    density: float = 1,
    smooth: float = 0.5,
    n_neighbors: Optional[int] = None,
    min_mass: float = 1,
    autoscale=True,
    stream_adjust=True,
    cutoff_percentile: float = 0.05,
    velocity_key: str = "velocity",
    self_transitions: bool = True,
    use_negative_cosines: bool = True,
    T_scale: float = 10,
    disable_scatter: bool = False,
    disable_cbar: bool = False,
    stream_kwargs: Optional[Dict[str, Any]] = {},
    scatter_kwargs: Optional[Dict[str, Any]] = {},
    cbar_kwargs: Optional[Dict] = {},
    mpl_kwargs: Optional[Dict[str, Any]] = {},
    return_axes: bool = False,
    save: Optional[bool] = False,
    rasterized: bool = True,
    png_dpi: Optional[float] = 500,
    svg_dpi: Optional[float] = 250,
    *args,
    **kwargs,
) -> Optional[Union[List[plt.Axes], None]]:
    """
    Generates velocity stream plots for single-cell data using the
    VelocityStreamPlot class.

    This function is a convenient wrapper around the VelocityStreamPlot
    class, allowing users to quickly generate and customize velocity stream
    plots without manually instantiating the class.

    Parameters
    ----------
    adata : anndata.AnnData
        The AnnData object containing the data to plot.
    ax : Optional[Union[plt.Axes, List[plt.Axes]]], optional
        Matplotlib axes object or list of axes objects on which to draw the plots. If None, a new figure and axes are created. **Default**: ``None``.
    c : str, optional
        Color for the scatter plot points. Can be a column name from `adata.obs` if coloring by a categorical variable. **Default**: ``"dodgerblue"``.
    cmap : Optional[Union[Dict, List, Tuple, str]], optional
        Colormap for the scatter plot points if `c` is a categorical variable. **Default**: ``"plasma_r"``.
    group_zorder : Optional[Dict], optional
        Z-order for groups in the scatter plot, allowing certain groups to be plotted on top of others. **Default**: ``None``.
    linewidth : float, optional
        Line width for the streamlines. **Default**: ``0.5``.
    stream_density : float, optional
        Density of the streamlines. Higher values create more densely packed streamlines. **Default**: ``2.5``.
    add_margin : float, optional
        Additional margin added around the plotted data, specified as a fraction of the data range. **Default**: ``0.1``.
    arrowsize : float, optional
        Size of the arrows in the stream plot. **Default**: ``1``.
    arrowstyle : str, optional
        Style of the arrows in the stream plot. **Default**: ``"-|>"``.
    maxlength : float, optional
        Maximum length of the arrows in the stream plot. **Default**: ``4``.
    integration_direction : str, optional
        Direction of integration for the streamlines, can be "forward", "backward", or "both". **Default**: ``"both"``.
    scatter_zorder : int, optional
        Z-order for scatter plot points, determining their layering. **Default**: ``101``.
    stream_zorder : int, optional
        Z-order for the streamlines, determining their layering. **Default**: ``201``.
    density : float, optional
        **Default**: ``1``.
    smooth : float, optional
        **Default**: ``0.5``.
    n_neighbors : Optional[int], optional
        **Default**: ``None``.
    min_mass : float, optional
        **Default**: ``1``.
    autoscale : bool, optional
        **Default**: ``True``.
    stream_adjust : bool, optional
        **Default**: ``True``.
    cutoff_percentile : float, optional
    velocity_key : str, optional
    self_transitions : bool, optional
    use_negative_cosines : bool, optional
    T_scale : float, optional
    disable_scatter : bool, optional
        If True, disables the scatter plot overlay on the stream plot. **Default**: ``False``.
    disable_cbar : bool, optional
        If True, disables the color bar for the scatter plot. Useful when `c` is numeric. **Default**: ``False``.
    stream_kwargs : Optional[Dict[str, Any]], optional
    scatter_kwargs : Optional[Dict[str, Any]], optional
    cbar_kwargs : Optional[Dict], optional
    mpl_kwargs : Optional[Dict[str, Any]], optional
        Additional keyword arguments for customizing the stream plot, scatter plot, color bar, and matplotlib figure, respectively.
    return_axes : bool, optional
        If True, returns the matplotlib axes with the generated plots. **Default**: ``False``.
    save : bool, optional
        If True, saves the generated plot to SVG and PNG formats. **Default**: ``False``.
    png_dpi : Optional[float], optional
        DPI settings for saving PNG images. **Default**: ``500``.
    svg_dpi : Optional[float], optional
        DPI settings for saving SVG images. **Default**: ``250``.

    Returns
    -------
    Optional[Union[List[plt.Axes], None]]
        A list of matplotlib axes with the generated plots, if ``return_axes == True``. Otherwise, returns ``None``.
    """

    init_kwargs = ABCParse.function_kwargs(VelocityStreamPlot.__init__, locals())
    call_kwargs = ABCParse.function_kwargs(VelocityStreamPlot.__call__, locals())
    velo_stream_plot = VelocityStreamPlot(**init_kwargs)
    axes = velo_stream_plot(**call_kwargs)

    if return_axes:
        return axes




__module_name__ = "__init__.py"
__doc__ = """I/O __init__ module. Sub-package of the main scdiffeq API."""
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["mvinyard.ai@gmail.com"])


# import functions accessed as sdq.io.<func>: --------------------------------------------
from ._data._read_h5ad import read_h5ad
from ._data._pickle_io import read_pickle, write_pickle



from ._model._project import Project
from ._model._version import Version
from ._model._checkpoint import Checkpoint
from ._model._hparams import HParams


from ._model._model_loader import load_diffeq, load_model, ModelLoader

# from ._logs import logs




import pandas as pd
import sklearn
import torch

from .. import base


class FateBiasDriftPriorMixIn(object):
    def __init__(self) -> None:
        super().__init__()
        
        
    def _configure_fate(
        self,
        graph,
        csv_path,
        t0_idx,
        fate_bias_multiplier = 1,
        undiff_key = "Undifferentiated",
    ) -> None:
        
        self.graph = graph
        self.fate_bias_multiplier = fate_bias_multiplier
        self.fate_df = pd.read_csv(csv_path, index_col=0)
        self.fate_df.index = t0_idx
        self._undiff_key = undiff_key

    def log_sinkhorn_divergence(self, sinkhorn_loss, t, stage, note=None):
        for i in range(len(t)):
            msg = f"sinkhorn_{t[i].item()}_{stage}"
            if note:
                msg = "_".join([note, msg])
            self.log(msg, sinkhorn_loss[i])

        return sinkhorn_loss.sum()
    
    def fate_accuracy(self, X_hat, batch_fate_idx):

        F_true = self.fate_df.loc[batch_fate_idx]
        self.X_hat = X_hat
        F_pred = self.graph(X_hat)
        F_pred.index = batch_fate_idx

        if F_pred.columns.unique().tolist() == [self._undiff_key]:
            return 0, 1

        univ_cols = [col for col in F_true.columns if col in F_pred.columns]
        F_true, F_pred = F_pred[univ_cols], F_true[univ_cols]
        acc_score = sklearn.metrics.accuracy_score(F_true.idxmax(1), F_pred.idxmax(1))
        acc_weight = 1 - acc_score
        return acc_score, acc_weight

    def step(self, batch, batch_idx=None, stage=None):
        
        # required
        batch = base.BatchProcessor(batch, batch_idx)
        X_hat, kl_div_loss = self.forward(batch.X0, batch.t)
        self.log(f"kl_div_{stage}", kl_div_loss.sum())
        
        sinkhorn_loss = self.compute_sinkhorn_divergence(
            batch.X, X_hat, batch.W, batch.W_hat,
        )
        self.log_sinkhorn_divergence(sinkhorn_loss=sinkhorn_loss, t=batch.t, stage=stage)
        acc_score, acc_weight = self.fate_accuracy(X_hat, batch.F_idx)
        
        acc_score = torch.Tensor([acc_score]).to(torch.float32)
        self.log(f"fate_acc_score_{stage}", acc_score)
            
        fate_weighted_sinkhorn_loss = sinkhorn_loss * acc_weight * self.fate_bias_multiplier

        self.log_sinkhorn_divergence(
            sinkhorn_loss=fate_weighted_sinkhorn_loss,
            t=batch.t,
            stage=stage,
            note="fate_weighted",
        ) 
        return sinkhorn_loss.sum() + fate_weighted_sinkhorn_loss.sum() + kl_div_loss.sum()




# -- import packages: ---------------------------------------------------------
import setuptools
import re
import os
import sys

# -- constants: --------------------------------------------------------------
name = "scdiffeq"

# -- fetch requirements packages: ---------------------------------------------
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

with open(f"{name}/__version__.py") as v:
    exec(v.read())


# -- run setup: ---------------------------------------------------------------
setuptools.setup(
    name=name,
    version=__version__,
    python_requires=">3.9.0",
    author="Michael E. Vinyard",
    author_email="mvinyard.ai@gmail.com",
    url="https://github.com/scDiffEq/scDiffEq",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    description="scDiffEq: modeling single-cell dynamics using neural differential equations.",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    license="MIT",
)




import abc
import numpy as np


class AnnDataIndexLookup(abc.ABC):
    def __init__(self):
        ...

    def _configure(self, adata):
        if not hasattr(self, "adata"):
            self.adata = adata.copy()

    @property
    @abc.abstractmethod
    def NAMES(self):
        ...

    def forward(self, idx):
        """Locate an individual cell index"""
        return np.where(idx == self.NAMES)[0]

    def __call__(self, adata, idx=[]) -> np.ndarray:
        self._configure(adata)
        return np.array([self.forward(ix) for ix in idx]).flatten()


class VarIndexLookUp(AnnDataIndexLookup):
    
    """
    At __call__, var_names are set and looked up.
    """
    
    def __init__(self, var_name_key: str = "gene_ids"):
        super().__init__()

        self._var_name_key = var_name_key

    def _set_var_names(self):
        self.adata.var_names = self.adata.var[self._var_name_key]

    def _configure(self, adata):
        if not hasattr(self, "adata"):
            self.adata = adata.copy()
            self._set_var_names()

    @property
    def NAMES(self):
        return self.adata.var_names


class ObsIndexLookUp(AnnDataIndexLookup):
    def __init__(self):
        ...

    @property
    def NAMES(self):
        return self.adata.obs_names



# -- import packages: ---------------------------------------------------------
import anndata
import ABCParse
import logging
import numpy as np
import pandas as pd
import logging

# -- configure logging: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- controlling class: -------------------------------------------------------
class CellFateAnnotation(ABCParse.ABCParse):
    """ """

    def __init__(
        self,
        time_key: str = "t",
        sim_key: str = "sim",
        silent: bool = False,
        *args,
        **kwargs,
    ) -> None:
        self.__parse__(locals())

    @property
    def _TIME(self) -> pd.Series:
        return self._adata_sim.obs[self._time_key]

    @property
    def _T_MAX(self) -> float:
        """Max time observed."""
        return self._TIME.max()

    @property
    def _FATE_VALUES(self) -> np.ndarray:
        return self._adata_sim.obs.loc[self._TIME == self._T_MAX][
            [self._sim_key, self._state_key]
        ].values

    def forward(self) -> None:
        mappable_fate_dict = {row[0]: row[1] for row in self._FATE_VALUES}
        self._adata_sim.obs[self._key_added] = self._adata_sim.obs[self._sim_key].map(
            mappable_fate_dict
        )
        if not self._silent:
            logger.info(f"Added fate annotation: adata_sim.obs['{self._key_added}']")
        self._count_fates()

    def _count_fates(self) -> None:
        """ """
        fate_obs = self._adata_sim.obs.loc[self._TIME == self._T_MAX]
        self._adata_sim.uns["fate_counts"] = (
            fate_obs[self._key_added].value_counts().to_dict()
        )

        if not self._silent:
            logger.info("Added fate counts: adata_sim.uns['fate_counts']")

    def __call__(
        self,
        adata_sim: anndata.AnnData,
        state_key: str,
        key_added: str = "fate",
        *args,
        **kwargs,
    ) -> None:
        self.__update__(locals())

        self.forward()


# -- API-facing function: -----------------------------------------------------
def annotate_cell_fate(
    adata_sim: anndata.AnnData,
    state_key: str = "state",
    key_added: str = "fate",
    time_key: str = "t",
    sim_key: str = "sim",
    silent: bool = False,
    *args,
    **kwargs,
) -> None:
    """Annotate the fate of a simulated cell trajectory, given annotated states.

    Parameters
    ----------
    adata_sim : anndata.AnnData
        Simulated AnnData object.

        state_key (str)
            Key in ``adata_sim.obs`` corresponding to annotated states.

        key_added (str)
            Key added to ``adata_sim.obs`` to indicate the fate of each trajectory.
            **Default**: "fate"

        time_key (str)
            Key in ``adata_sim.obs`` corresponding to time.

        sim_key (str)
            Key in ``adata_sim.obs`` corresponding to each individual simulation.

        silent (bool)
            If ``True``, suppresses informational messages.

    Returns
    -------
    None
    """

    fate_annot = CellFateAnnotation(time_key=time_key, sim_key=sim_key, silent=silent)
    fate_annot(adata_sim=adata_sim, state_key=state_key, key_added=key_added)



# -- import packages: ---------------------------------------------------------
import ABCParse
import adata_query
import anndata
import logging
import numpy as np
import scipy.sparse
import warnings

# -- import local dependencies: -----------------------------------------------
from ._norm import L2Norm
from ._get_neighbor_indices import get_neighbor_indices
from ._get_iterative_indices import get_iterative_indices

# -- set type hints: ----------------------------------------------------------
from typing import Optional

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- supporting operational class: --------------------------------------------
class CosineCorrelation(ABCParse.ABCParse):
    def __init__(self, *args, **kwargs) -> None:
        """ """
        self.__parse__(locals())

        self._L2Norm = L2Norm()

    def _mean_subtraction(self, dX: np.ndarray) -> np.ndarray:
        return dX - dX.mean(-1)[:, None]

    def _Vi_norm(self, Vi: np.ndarray) -> np.ndarray:
        return self._L2Norm(Vi, axis=0)

    def forward(self, dX: np.ndarray, Vi: np.ndarray) -> np.ndarray:
        dx = self._mean_subtraction(dX)
        Vi_norm = self._Vi_norm(Vi)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if Vi_norm == 0:
                return np.zeros(dx.shape[0])
            return (
                np.einsum("ij, j", dx, Vi)
                / (self._L2Norm(dx, axis=1) * Vi_norm)[None, :]
            )

    def __call__(self, dX: np.ndarray, Vi: np.ndarray, *args, **kwargs) -> np.ndarray:
        """ """
        self.__update__(locals())

        return self.forward(dX, Vi)


# -- operational class: -----------------------------------------------------------------
class ComputeCosines(ABCParse.ABCParse):
    def __init__(
        self,
        state_key="X",
        velocity_key: str = "X_drift",
        n_pcs: Optional[int] = None,
        split_negative: bool = True,
        distances_key: str = "distances", # NEW
        *args,
        **kwargs,
    ) -> None:
        self.__parse__(locals())

        self._L2Norm = L2Norm()
        self._cosine_correlation = CosineCorrelation()

    def _initialize_results(self) -> None:
        self._VALS, self._ROWS, self._COLS = [], [], []

    def _l2_norm(self, input):
        return self._L2Norm(input)

    @property
    def X(self):
        if not hasattr(self, "_X"):
            self._x = adata_query.fetch(self._adata, key=self._state_key, torch=False)
            if not self._n_pcs is None:
                self._x = self._x[:, : self._n_pcs]
        return self._x

    @property
    def V(self):
        if not hasattr(self, "_V"):
            self._v = adata_query.fetch(
                self._adata, key=self._velocity_key, torch=False
            )
            if not self._n_pcs is None:
                self._v = self._v[:, : self._n_pcs]
        return self._v

    @property
    def obs_idx(self):
        return range(len(self._adata))

    @property
    def n_neighbors(self) -> int:  # NEW
        if not hasattr(self, "_n_neighbors"):
            n_neighbor_param = self._adata.uns["neighbors"]["params"]["n_neighbors"]
            if isinstance(n_neighbor_param, int):
                self._n_neighbors = n_neighbor_param
            else:
                self._n_neighbors = n_neighbor_param[0]
        return self._n_neighbors

    @property
    def nn_idx(self):
        if not hasattr(self, "_nn_idx"):
            self._nn_idx = get_neighbor_indices(
                adata=self._adata,
                n_neighbors=self.n_neighbors,
                distances_key=self._distances_key,
            )
        return self._nn_idx

    def _contains_non_zero(self, obs_id: int) -> bool:
        return np.any(np.array([self.V[obs_id].min(), self.V[obs_id].max()]) != 0)

    def forward(self, obs_id) -> None:
        if self._contains_non_zero(obs_id):
            iter_nn_idx = get_iterative_indices(self.nn_idx, obs_id, 2, None)
            dX = self.X[iter_nn_idx] - self.X[obs_id, None]
            corr = self._cosine_correlation(dX, self.V[obs_id])
            self._VALS.extend(corr)
            self._ROWS.extend(np.ones(len(iter_nn_idx)) * obs_id)
            self._COLS.extend(iter_nn_idx)

    def _replace_nans(self, vals):
        vals = np.hstack(vals)
        vals[np.isnan(vals)] = 0
        return vals

    def _perform_split_negative(self, graph):
        graph_neg = graph.copy()

        graph.data = np.clip(graph.data, 0, 1)
        graph_neg.data = np.clip(graph_neg.data, -1, 0)

        graph.eliminate_zeros()
        graph_neg.eliminate_zeros()

        return graph.tocsr(), graph_neg.tocsr()

    def _assemble_graph_as_csr(self):
        self._VALS = self._replace_nans(self._VALS)

        graph = scipy.sparse.coo_matrix(
            (self._VALS, (self._ROWS, self._COLS)),
            shape=(len(self.obs_idx), len(self.obs_idx)),
        )
        if self._split_negative:
            return self._perform_split_negative(graph)
        return graph

    def _update_adata(
        self, graph: scipy.sparse.csr_matrix, graph_neg: scipy.sparse.csr_matrix
    ) -> None:

        pos_key = f"{self._velocity_key_added}_graph"
        neg_key = f"{self._velocity_key_added}_graph_neg"

        for key, obj in zip([pos_key, neg_key], [graph, graph_neg]):
            if key in self._adata.obsp:
                self._adata.obsp[key] = obj
                logger.info(f"Updated: adata.obsp['{key}']")
            else:
                self._adata.obsp[key] = obj
                logger.info(f"Added: adata.obsp['{key}']")

    def __call__(
        self,
        adata: anndata.AnnData,
        velocity_key_added: str = "velocity",
        *args,
        **kargs,
    ) -> None:
        """ """
        self.__update__(locals())

        self._initialize_results()

        for obs_id in self.obs_idx:
            self.forward(obs_id)

        self._update_adata(*self._assemble_graph_as_csr())




# -- import packages: ---------------------------------------------------------
import ABCParse
import lightning
import os

# -- import local dependencies: -----------------------------------------------
from .. import utils, callbacks as _callbacks

# -- set type hints: ----------------------------------------------------------
from typing import List

# -- supporting class: --------------------------------------------------------
class InterTrainerEpochCounter(lightning.pytorch.callbacks.Callback):
    def __init__(self):
        ...
#         self.COMPLETED_EPOCHS = 0

    def on_train_epoch_end(self, trainer, pl_module, *args, **kwargs):

        pl_module.COMPLETED_EPOCHS += 1
        ce = pl_module.COMPLETED_EPOCHS

        
# -- operational class: -------------------------------------------------------
class LightningCallbacksConfiguration(ABCParse.ABCParse):
    def __init__(self):
        super().__init__()
        
        self.cbs = []

    @property
    def BuiltInCallbacks(self):

        return [
            lightning.pytorch.callbacks.ModelCheckpoint(
                every_n_epochs=self._every_n_epochs,
                save_on_train_epoch_end=True,
                save_top_k=self._save_top_k,
                save_last=self._save_last,
                monitor=self._monitor,
            ),
            InterTrainerEpochCounter(),
            _callbacks.ModelLogging(),
        ]
    
    @property
    def BuiltInPreTrainCallbacks(self):

        return [
            ModelCheckpoint(
                every_n_epochs=self._every_n_epochs,
                save_on_train_epoch_end=True,
                save_top_k=self._save_top_k,
                save_last=self._save_last,
                monitor=self._monitor,
            ),
            InterTrainerEpochCounter(),
        ]
    
    @property
    def DEPLOYED_CALLBACKS(self):
        return self.cbs + self.BuiltInCallbacks
    
    def _update_callbacks(self, callbacks: List[lightning.pytorch.callbacks.Callback]) -> None:
        
        [self.cbs.append(cb) for cb in callbacks]
        
        if self._monitor_hardware:
            self.cbs.append(_callbacks.MemoryMonitor())
        if self._retain_test_gradients:
            self.cbs.append(_callbacks.GradientPotentialTest())

    def __call__(
        self,
        version,
        stage,
        monitor_hardware: bool = False,
        viz_frequency = 1,
        model_name="scDiffEq_model",
        working_dir=os.getcwd(),
        train_version=0,
        pretrain_version=0,
        callbacks=[],
        ckpt_frequency=10,
        keep_ckpts=-1,
        swa_lrs=1e-5,
        monitor=None,
        retain_test_gradients=False,
        save_last=True,
    ):
        
        self.__parse__(locals(), ignore=["callbacks"])
        
        self._every_n_epochs = ckpt_frequency
        self._save_top_k = keep_ckpts
        
        self._update_callbacks(callbacks=callbacks)
        
        return self.DEPLOYED_CALLBACKS



import ABCParse
import pathlib
import anndata
import scanpy as sc
import os
import sklearn
import umap
import numpy as np
import web_files


from .. import io


class HumanHematpoiesisDataHandler(ABCParse.ABCParse):
    HTTPS = "https://www.dropbox.com/s/8m8n6fj8yn1ryjd/hsc_all_combined_all_layers.h5ad?dl=1"
    RAW_FNAME = "_hsc_all_combined_all_layers.h5ad"
    PROCESSED_FNAME = "human_hematopoiesis.preprocessed.h5ad"

    def __init__(self, data_dir=os.getcwd(), *args, **kwargs):
        self.__parse__(locals())

    @property
    def _scdiffeq_parent_data_dir(self):
        path = pathlib.Path(self._data_dir).joinpath("scdiffeq_data")
        if not path.exists():
            path.mkdir()
        return path

    @property
    def data_dir(self):
        path = self._scdiffeq_parent_data_dir.joinpath("human_hematopoiesis")
        if not path.exists():
            path.mkdir()
        return path

    @property
    def raw_h5ad_path(self) -> pathlib.Path:
        return self.data_dir.joinpath(self.RAW_FNAME)

    @property
    def processed_h5ad_path(self) -> pathlib.Path:
        return self.data_dir.joinpath(self.PROCESSED_FNAME)

    def download_raw(self):
        web_file = web_files.WebFile(
            http_address=self.HTTPS, local_path=str(self.raw_h5ad_path)
        )
        web_file.download()

    @property
    def raw_adata(self):
        if not hasattr(self, "_raw_adata"):
            if not self.raw_h5ad_path.exists():
                self.download_raw()
            self._raw_adata = anndata.read_h5ad(self.raw_h5ad_path)
            self._raw_idx = self._raw_adata.obs.index
        return self._raw_adata

    @property
    def pp_adata(self):
        if not hasattr(self, "_pp_adata"):
            if not self.processed_h5ad_path.exists():
                self._pp_adata = self.get()
            else:
                self._pp_adata = anndata.read_h5ad(self.processed_h5ad_path)
        return self._pp_adata

    def seurat_preprocessing(self, adata):

        adata = adata.copy()

        adata.obs["time"] = 3
        adata.obs.loc[adata.obs.batch == "JL_10", "time"] = 5
        adata.obs.time.value_counts()
        adata.obs_names_make_unique()
        adata.obs["nGenes"] = (adata.X > 0).sum(1)
        adata.obs["UMI"] = (adata.X).sum(1)
        adata.layers["X_orig"] = adata.X
        adata.X = adata.layers["labeled_TC"].copy()
        adata.X[adata.obs.batch == "JL_10", :] *= 3 / 5
        adata.obs.rename({"time": "_time"}, axis=1, inplace=True)
        t_map = {"JL_10": 4, "JL12_0": 7, "JL12_1": 7}
        adata.obs["t"] = adata.obs["batch"].map(t_map)
        sc.pp.recipe_seurat(adata)

        return adata

    def _match_and_filter_on_idx(self, adata):
        """"""

        HTTPS = "https://www.dropbox.com/s/n9mx9trv1h78q0r/hematopoiesis_v1.h5ad?dl=1"
        _secondary_h5ad = self.data_dir.joinpath("_dynamo_hematopoiesis_v1.h5ad")
        if not _secondary_h5ad.exists():
            web_file = web_files.WebFile(
                http_address=HTTPS, local_path=str(_secondary_h5ad)
            )
            web_file.download()

        adata_dyn = anndata.read_h5ad(_secondary_h5ad)

        match_idx = np.where(adata.obs.index.isin(adata_dyn.obs.index))
        adata = adata[match_idx].copy()
        adata.obs["cell_type"] = adata_dyn[adata.obs.index].obs["cell_type"].values

        return adata

    def dimension_reduction(self, adata: anndata.AnnData, skip_scaling: bool = True):

        adata = adata.copy()

        self.SCALER_MODEL = sklearn.preprocessing.StandardScaler()
        self.PCA_MODEL = sklearn.decomposition.PCA(n_components=50)
        self.UMAP_MODEL = umap.UMAP(
            n_components=2,
            n_neighbors=25,
            random_state=1,
            min_dist=0.5,
        )

        if skip_scaling:
            adata.obsm["X_pca"] = self.PCA_MODEL.fit_transform(adata.X)
        else:
            adata.layers["X_scaled"] = self.SCALER_MODEL.fit_transform(adata.X)
            adata.obsm["X_pca"] = self.PCA_MODEL.fit_transform(adata.layers["X_scaled"])
            io.write_pickle(
                self.SCALER_MODEL,
                self.data_dir.joinpath("human_hematopoiesis.scaler_model.pkl"),
            )

        adata.obsm["X_umap"] = self.UMAP_MODEL.fit_transform(
            adata.obsm["X_pca"][:, :10]
        )

        io.write_pickle(
            self.PCA_MODEL, self.data_dir.joinpath("human_hematopoiesis.pca_model.pkl")
        )
        io.write_pickle(
            self.UMAP_MODEL,
            self.data_dir.joinpath("human_hematopoiesis.umap_model.pkl"),
        )

        return self._match_and_filter_on_idx(adata)

    def get(self):

        adata_pp = self.seurat_preprocessing(self.raw_adata)
        adata_pp = self.dimension_reduction(adata_pp)
        adata_pp.write_h5ad(self.processed_h5ad_path)
        return adata_pp


def human_hematopoiesis(data_dir=os.getcwd()) -> anndata.AnnData:
    """Human hematopoiesis dataset from Dynamo.

    Parameters
    ----------
    data_dir : str, default=os.getcwd()
        Path to the directory where the data will be saved.

    Returns
    -------
    anndata.AnnData
        Processed AnnData object.
    """
    data_handler = HumanHematpoiesisDataHandler(data_dir=data_dir)
    return data_handler.pp_adata




# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata


# -- set typing: --------------------------------------------------------------
from typing import Optional


# -- operational class: -------------------------------------------------------
class TimeKeyConfiguration(ABCParse.ABCParse):
    def __init__(self, *args, **kwargs):
        self.__parse__(locals())

    @property
    def _AUTO_DETECTED_TIME_COLS(self):
        return [
            "t",
            "Time point",
            "time",
            "time_pt",
            "t_info",
            "time_info",
        ]

    @property
    def _OBS_COLS(self):
        return self._adata.obs.columns

    @property
    def _PASSED(self) -> bool:
        return hasattr(self, "_time_key")

    @property
    def _PASSED_VALID(self) -> bool:
        if self._PASSED:
            return self._time_key in self._OBS_COLS

    @property
    def _DETECTED(self) -> bool:
        return self._OBS_COLS[
            [col in self._AUTO_DETECTED_TIME_COLS for col in self._OBS_COLS]
        ].tolist()

    @property
    def _DETECTED_VALID(self) -> bool:
        return len(self._DETECTED) == 1

    @property
    def _VALID(self):
        return any([self._PASSED_VALID, self._DETECTED_VALID])

    @property
    def _KeyNotFoundError(self):
        raise KeyError(f"Passed `time_key`: {self._time_key} not found in adata.obs")

    @property
    def _MultipleKeysFoundError(self):
        msg = f"More than one time column inferred: {self._DETECTED}. Specify the desired time column in adata.obs."
        raise KeyError(msg)

    def __call__(
        self,
        adata: anndata.AnnData,
        time_key: Optional[str] = None,
        default_time_key: str = "t",
        *args,
        **kwargs,
    ) -> str:

        self.__update__(locals())

        if self._PASSED:
            if self._PASSED_VALID:
                return self._time_key
            raise self._KeyNotFoundError()

        if self._DETECTED:
            if self._DETECTED_VALID:
                return self._DETECTED[0]
            raise self._MultipleKeysFoundError()

        return self._default_time_key


# -- model-facing function: ---------------------------------------------------
def configure_time_key(
    adata: anndata.AnnData,
    time_key: Optional[str] = None,
) -> str:
    """
    Configure time key

    Parameters
    ----------
    adata: anndata.AnnData

    time_key: Optional[str], default = None

    Returns
    -------
    time_key: str
    """
    time_key_config = TimeKeyConfiguration()
    return time_key_config(adata=adata, time_key=time_key)



import ABCParse
import datetime
import glob
import logging
import os
import pathlib

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- cls: -------------------------------------------------------------------
class scDiffEqLogger(ABCParse.ABCParse):
    """
    While Lightning uses automatic logging, we need something one step removed from this to take full
    advantage of their setup within the constraints of our model.
    """

    def __init__(
        self,
        model_name="scDiffEq_model",
        ckpt_path=None,
        working_dir=os.getcwd(),
    ) -> None:

        self.__parse__(locals(), public=[None])
        self.creation_count = 0

    @property
    def _MODEL_NAME(self):
        return self._model_name

    @property
    def _WORKING_DIR(self):
        return self._working_dir

    @property
    def _EXISTING_VERSIONS(self):
        return glob.glob(self.PARENT_MODEL_OUTDIR + "/version*")

    @property
    def PARENT_MODEL_OUTDIR(self):
        return os.path.join(self._WORKING_DIR, self._MODEL_NAME)

    @property
    def LOG_PATH(self):
        return os.path.join(self.PARENT_MODEL_OUTDIR, "scDiffEq.log")

    @property
    def VERSIONED_MODEL_OUTDIR(self):
        if not self._ckpt_path is None:
            if self._PASSED_CKPT_MATCHES_MODEL_PARENT_DIR:
                return self.VERSION_FROM_CKPT

        return os.path.join(
            self.PARENT_MODEL_OUTDIR,
            "version_{}".format(len(self._EXISTING_VERSIONS)),
        )

    def _configure_parent_model_outdir(self):
        if not os.path.exists(self.PARENT_MODEL_OUTDIR):
            os.mkdir(self.PARENT_MODEL_OUTDIR)
            line = "".join(
                [
                    "\n ------------- scDiffEq -------------\n\n\n",
                    f" ---- {datetime.datetime.now()} ----\n\n",
                ]
            )

            f = open(self.LOG_PATH, mode="w")
            f.write(line)
            f.close()

    def _configure_versioned_model_outdir(self):

        if not os.path.exists(self.VERSIONED_MODEL_OUTDIR):
            if not self.creation_count:
                os.mkdir(self.VERSIONED_MODEL_OUTDIR)
            f = open(self.LOG_PATH, mode="a")
            v_path = self.VERSIONED_MODEL_OUTDIR
            v = os.path.basename(v_path)
            line = f"\t{v}\t{datetime.datetime.now()}\t{v_path}\n"
            f.write(line)
            f.close()
            self.creation_count += 1

        elif self._PASSED_CKPT_MATCHES_MODEL_PARENT_DIR:
            if len(self.CKPT_PATH) < 65:
                logger.info(f"Loading from checkpoint: {self.CKPT_PATH}")
            else:
                logger.info(f"Loading from checkpoint:\n\t{self.CKPT_PATH}")
        else:
            logger.info(f"Directory: {self.VERSIONED_MODEL_OUTDIR} already exists!")

    # -- checkpoint loading: ------------------------------------------------
    @property
    def CKPT_PATH(self):
        return pathlib.Path(self._ckpt_path).absolute().as_posix()

    @property
    def VERSION_FROM_CKPT(self):
        return os.path.dirname(self.CKPT_PATH.split("_logs/version_")[0])

    @property
    def _PASSED_CKPT_MATCHES_MODEL_PARENT_DIR(self) -> bool:
        compare = [self.CKPT_PATH, self.PARENT_MODEL_OUTDIR]
        common = os.path.commonprefix(compare)

        return common == self.PARENT_MODEL_OUTDIR

    def __call__(self) -> None:

        self._configure_parent_model_outdir()
        self._configure_versioned_model_outdir()



# -- import packages: --------------------------------------------------------------------
from lightning import Callback
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import vinplots
import os


# -- Helper functions: -------------------------------------------------------------------
def _epoch_regex_mean(epoch_df, regex):
    return epoch_df.filter(regex=regex).dropna().mean(0)


# -- Plotting: ---------------------------------------------------------------------------
class LossPlotter(vinplots.Plot):
    def __init__(self, time_unit="", x_label="Epoch", y_label="Wasserstein Distance"):

        self.build()
        self.time_unit = time_unit
        self._colors = np.array(list(vinplots.colors.pbmc3k.values()))
        # self.colors = self._colors[[4, 6, 7]]
        cc = vinplots.colors.BuOr.colors
        self.colors = cc[:int(len(cc)/2)][::3]
        self.x_label = x_label
        self.y_label = y_label
        self.y_min = 0
        self.y_max = 300
        self.legend_fontsize = "small"
        self.fit_colors = {"train": self._colors[0], "val": self._colors[2]}

    def build(self):

        self.construct(nplots=3, ncols=3, wspace=0.1, figsize=0.8)
        self.modify_spines(ax="all", spines_to_delete=["top", "right"])
        self.axes = self.linearize()

    def _format_ax(self, ax, title):

        ax.tick_params(axis="both", which="both", labelsize=6)
        ax.set_xlabel(self.x_label, fontsize=8)
        ax.set_ylabel(self.y_label, fontsize=8)
#         ax.set_ylim(self.y_min, self.y_max)
        ax.set_title(title, fontsize=10)
        ax.grid(True, zorder=0, alpha=0.2)

    def _plot_ax(self, ax, loss_df, fit=False, title=None):

        if not fit:
            iterable = loss_df.columns[1:-1]
        else:
            iterable = loss_df.columns
        
        for n, i in enumerate(iterable):
            if isinstance(i, int):
                label = "d{}".format(i)
            else:
                label = i
            ax.plot(loss_df[i], c=self.colors[n], label=label)
            if n == (len(iterable)-1):
                ax.legend(edgecolor="w", fontsize=self.legend_fontsize)
        self._format_ax(ax, title)

        
    def forward(self, loss, savepath):

        self._plot_ax(self.axes[0], loss.train, fit=False, title="Training Loss")
        self._plot_ax(self.axes[1], loss.val, fit=False, title="Validation Loss")
        self._plot_ax(self.axes[2], loss.fit, fit=True,  title="Fitting Loss")

        if savepath:
            plt.savefig(savepath + ".svg")
            plt.savefig(savepath + ".png")


# -- Get loss: ---------------------------------------------------------------------------
class LossReader:
    def __init__(self, trainer):

        self._log_dir = trainer.logger.log_dir
        self._metrics_path = os.path.join(self._log_dir, "metrics.csv")
    
    @property
    def metrics_csv_exists(self):
        return os.path.exists(self._metrics_path)
    
    @property
    def metrics_df(self):
        if self.metrics_csv_exists:
            return pd.read_csv(self._metrics_path)
    
    def _adjust_colnames(self, filtered_loss_df):

        cols = [int(col.split("_")[1]) for col in filtered_loss_df.columns]
        filtered_loss_df.columns = cols
        filtered_loss_df["sum"] = filtered_loss_df.sum(1).values

        return filtered_loss_df

    def filter_loss_df(self, regex, groupby="epoch"):

        filtered_loss_df = (
            self.metrics_df.groupby(groupby).apply(_epoch_regex_mean, regex=regex).copy()
        )
        return self._adjust_colnames(filtered_loss_df).dropna()

    @property
    def train(self) -> pd.DataFrame:
        return self.filter_loss_df(regex="train")

    @property
    def val(self) -> pd.DataFrame:
        return self.filter_loss_df(regex="val")

    @property
    def fit(self) -> pd.DataFrame:
        return pd.DataFrame({"train": self.train["sum"], "val": self.val["sum"]})

    def plot(self, savepath=None):

        if self.metrics_csv_exists:
            self.plotter = LossPlotter()
            self.plotter.forward(self, savepath=savepath)


# -- Callback: ---------------------------------------------------------------------------
class LossAccounting(Callback):
    def __init__(self, save_img=True):
        
        self.save_img = save_img
        self.savepath = None
    
    def _forward(self, trainer):
        
        loss = LossReader(trainer)
        if loss.metrics_csv_exists:
            if self.save_img:
                self.savepath = os.path.join(trainer.logger.log_dir, "model_fit_loss")
            loss.plot(savepath=self.savepath)
        
    def on_validation_epoch_end(self, trainer, model):
        
        self._forward(trainer)
        plt.close();
        
    def on_fit_end(self, trainer, model):
        
        self._forward(trainer)



from ._anndata_inspector import AnnDataInspector
from ._function_kwargs import extract_func_kwargs

import annoyance
import pandas as pd
import torch
import ABCParse
import logging

# -- configure logging: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

NoneType = type(None)


class kNNGraphQuery(ABCParse.ABCParse):
    def __init__(self, adata, use_key, *args, **kwargs):

        self.__parse__(locals())
        self._IGNORE += ["X_hat", "query_t"]

        self._configure_graph()

    def configure_used_layer(self):
        inspector = AnnDataInspector(self.adata)
        if self.use_key in inspector.layers:
            self.adata.obsm[self.use_key] = self.adata.layers[self.use_key]

    @property
    def _GRAPH_KWARGS(self):
        return extract_func_kwargs(annoyance.kNN, self._PARAMS)

    def _configure_graph(self):

        self.configure_used_layer()
        self.Graph = annoyance.kNN(**self._GRAPH_KWARGS)
        self.Graph.build()

    def _format_input_shape(self, X_hat, query_t):
        """"""
        if not isinstance(query_t, NoneType):
            return X_hat[query_t][None, :]
        return X_hat

    def _format_input_device(self, X_hat: torch.Tensor):

        if X_hat.device != "cpu":
            return X_hat
        return X_hat.detach().cpu().numpy()

    def _format_input(self, X_hat, query_t):
        return self._format_input_device(self._format_input_shape(X_hat, query_t))

    def _count(self, X_nn):
        return (
            self.adata[X_nn.flatten().astype(str)]
            .obs[self.annot_key]
            .values.reshape(-1, self.Graph._n_neighbors)
        )

    def _query(self, X_hat):
        return self._count(self.Graph.query(X_hat))

    def _fate_df(self, x_lab):
        return pd.DataFrame([x_lab[0].value_counts() for i in range(len(x_lab))])

    def __call__(self, X_hat: torch.Tensor, annot_key: str, query_t: int = -1):
        """
        X_hat: torch.Tensor of shape: [t, n_sim, n_dim]
        """
        self.__update__(locals())

        X_hat = self._format_input(X_hat, query_t)
        return pd.DataFrame([self._fate_df(self._query(X)).idxmax(1) for X in X_hat]).T
